@@ -10,8 +10,8 @@ use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
 use crate::{
     apply_fn_cache, deep_map_op_context_preorder_stoppable, deep_map_op_pre_new_children,
-    deep_map_unwrap, visit_circuit_unwrap, ArrayConstant, Circuit, CircuitNode, CircuitRc,
-    HashBytes, Index,
+    deep_map_unwrap, visit_circuit_non_free, visit_circuit_unwrap, Array, Circuit, CircuitNode,
+    CircuitRc, HashBytes, Index,
 };
 
 #[pyfunction]
@@ -19,9 +19,9 @@ pub fn cast_circuit(circ: CircuitRc, device_dtype: &TorchDeviceDtype) -> Circuit
     deep_map_op_pre_new_children(
         circ.clone(),
         |circ, children: &Vec<CircuitRc>| match &**circ {
-            Circuit::ArrayConstant(node) => {
+            Circuit::Array(node) => {
                 if &TorchDeviceDtype::from_tensor(&node.value) != device_dtype {
-                    Some(ArrayConstant::nrc(
+                    Some(Array::nrc(
                         device_dtype.cast_tensor(node.value.clone()),
                         node.name_cloned(),
                     ))
@@ -77,14 +77,27 @@ pub fn count_nodes(circuit: CircuitRc) -> usize {
     result
 }
 
-pub fn hash_to_node(circuit: CircuitRc) -> HashMap<HashBytes, CircuitRc> {
+pub fn hash_to_node_non_free(
+    circuit: CircuitRc,
+    recur_into_free: bool,
+) -> HashMap<HashBytes, CircuitRc> {
     let mut result: HashMap<HashBytes, CircuitRc> = HashMap::default();
-    visit_circuit_unwrap(circuit, &mut |x: CircuitRc| {
-        result.insert(x.info().hash, x.clone());
-    });
+    visit_circuit_non_free(
+        circuit,
+        &mut |x: CircuitRc| {
+            result.insert(x.info().hash, x.clone());
+            Ok(())
+        },
+        recur_into_free,
+    )
+    .unwrap();
     result
 }
 
+pub fn hash_to_node(circuit: CircuitRc) -> HashMap<HashBytes, CircuitRc> {
+    hash_to_node_non_free(circuit, true)
+}
+
 pub fn subcircuits(circuit: CircuitRc) -> HashSet<CircuitRc> {
     let mut result: HashSet<CircuitRc> = HashSet::default();
     visit_circuit_unwrap(circuit, &mut |x: CircuitRc| {
@@ -115,7 +128,7 @@ pub fn total_flops_cached(circuit: CircuitRc, cache: &mut HashMap<HashBytes, Big
 pub fn total_arrayconstant_size(circuit: CircuitRc) -> BigUint {
     let mut result: BigUint = BigUint::from(0usize);
     visit_circuit_unwrap(circuit, &mut |x: CircuitRc| {
-        if let Circuit::ArrayConstant(x) = &**x {
+        if let Circuit::Array(x) = &**x {
             result += x.info().numel();
         }
     });
@@ -126,7 +139,7 @@ pub fn total_arrayconstant_size(circuit: CircuitRc) -> BigUint {
 pub fn sum_of_node_sizes(circuit: CircuitRc) -> BigUint {
     let mut result: BigUint = BigUint::from(0usize);
     visit_circuit_unwrap(circuit, &mut |x: CircuitRc| {
-        if !matches!(&**x, Circuit::ArrayConstant(_)) {
+        if !matches!(&**x, Circuit::Array(_)) {
             result += x.info().numel();
         }
     });
@@ -289,7 +302,7 @@ where
 #[pyfunction]
 pub fn replace_all_randn_seeded(circuit: CircuitRc) -> CircuitRc {
     deep_map_unwrap(circuit, |node| match &**node {
-        Circuit::ArrayConstant(ac) => ArrayConstant::randn_full(
+        Circuit::Array(ac) => Array::randn_full(
             ac.info().shape.clone(),
             ac.name_cloned(),
             ac.info().device_dtype.clone(),
@@ -320,29 +333,6 @@ pub fn replace_all_randn_seeded(circuit: CircuitRc) -> CircuitRc {
     })
 }
 
-pub fn has_leaf_symbols(circuit: CircuitRc, allowed: &HashSet<HashBytes>) -> bool {
-    let mut result = false;
-    crate::visit_circuit_unwrap(circuit, |circuit| {
-        if circuit.as_symbol().is_some() && !allowed.contains(&circuit.info().hash) {
-            println!(
-                "found symbol {:?} allowed {:?}",
-                circuit.info().hash,
-                allowed
-            );
-            result = true;
-        }
-        if let Circuit::ModuleNode(mn) = &**circuit {
-            let inner_allowed: HashSet<HashBytes> = allowed
-                .iter()
-                .cloned()
-                .chain(mn.spec.input_specs.iter().map(|s| s.symbol.info().hash))
-                .collect();
-            result = result || has_leaf_symbols(mn.spec.spec_circuit.clone(), &inner_allowed);
-        }
-    });
-    result
-}
-
 /// Replaces child nodes recursively with a mapping
 /// In pre-order so that high level nodes are replaced first
 pub fn replace_nodes(circuit: CircuitRc, map: &HashMap<HashBytes, CircuitRc>) -> CircuitRc {
