@@ -28,10 +28,16 @@ use rr_util::{
     },
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
+use uuid::uuid;
 
 use crate::{
-    check_canon_axes, circuit_node_auto_impl, circuit_node_extra_impl, new_rc_unwrap, prelude::*,
-    CachedCircuitInfo, ConstructError, PyCircuitBase, ScalarConstant, Shape,
+    check_canon_axes, circuit_node_auto_impl, circuit_node_extra_impl,
+    circuit_node_private::{
+        CircuitNodeComputeInfoImpl, CircuitNodeHashItems, CircuitNodeHashWithChildren,
+    },
+    new_rc_unwrap,
+    prelude::*,
+    CachedCircuitInfo, CircuitNodeSelfOnlyHash, ConstructError, PyCircuitBase, Scalar, Shape,
 };
 
 /// our Einsum supports diag output
@@ -48,25 +54,34 @@ pub struct Einsum {
 
 circuit_node_extra_impl!(Einsum);
 
-impl CircuitNode for Einsum {
-    circuit_node_auto_impl!("ed15422c-7c02-40c2-a3c2-e9224514d063");
-
+impl CircuitNodeComputeInfoImpl for Einsum {
     fn compute_shape(&self) -> Shape {
         let map = self.shape_map().unwrap();
         self.out_axes.iter().map(|a| *map.get(a).unwrap()).collect()
     }
-
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
+}
+impl CircuitNodeHashWithChildren for Einsum {
+    fn compute_hash_non_name(&self, hasher: &mut blake3::Hasher) {
         hasher.update(&self.out_axes);
         for (c, axes) in &self.args {
             // no need to delimit because we have hash each loop
             hasher.update(&c.info().hash);
             hasher.update(axes);
         }
-
-        hasher
     }
+}
+impl CircuitNodeSelfOnlyHash for Einsum {
+    fn compute_self_only_hash(&self, hasher: &mut blake3::Hasher) {
+        hasher.update(&self.out_axes);
+        for (_c, axes) in &self.args {
+            hasher.update(uuid!("3749ab7f-4f4c-4fdb-a268-af9c084e70ef").as_bytes());
+            hasher.update(axes);
+        }
+    }
+}
+
+impl CircuitNode for Einsum {
+    circuit_node_auto_impl!("ed15422c-7c02-40c2-a3c2-e9224514d063");
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         let out_inv: HashMap<AxisInt, usize> = hashmap_collect_except_duplicates(
@@ -152,19 +167,21 @@ impl Einsum {
         for (circ, axes) in &self.args {
             if circ.info().rank() != axes.len() {
                 bail!(ConstructError::EinsumLenShapeDifferentFromAxes {
-                    circuit_name: circ.name_cloned(),
-                    circuit_len_shape: circ.info().rank(),
                     len_axes: axes.len(),
+                    child_len_shape: circ.info().rank(),
+                    einsum_name: self.name_cloned(),
+                    child_circuit: circ.clone(),
                 });
             }
-            for (&circuit_shape, axis) in circ.info().shape.iter().zip(axes) {
-                let existing_shape = *out.entry(*axis).or_insert(circuit_shape);
-                if existing_shape != circuit_shape {
-                    bail!(ConstructError::EinsumShapeDifferent {
-                        circuit_name: circ.name_cloned(),
-                        circuit_shape,
+            for (&circuit_size, axis) in circ.info().shape.iter().zip(axes) {
+                let existing_size = *out.entry(*axis).or_insert(circuit_size);
+                if existing_size != circuit_size {
+                    bail!(ConstructError::EinsumAxisSizeDifferent {
+                        new_size: circuit_size.into(),
+                        existing_size: existing_size.into(),
                         axis: *axis as usize,
-                        existing_shape,
+                        einsum_name: self.name_cloned(),
+                        child_circuit: circ.clone(),
                     });
                 }
             }
@@ -172,6 +189,7 @@ impl Einsum {
 
         Ok(out)
     }
+
     #[apply(new_rc_unwrap)]
     pub fn try_new(args: EinsumArgs, out_axes: EinsumAxes, name: Option<String>) -> Result<Self> {
         // TODO: reduce axes nums!
@@ -187,7 +205,12 @@ impl Einsum {
 
         for a in &out.out_axes {
             if !map.contains_key(a) {
-                return Err(ConstructError::EinsumOutputNotSubset {}.into());
+                return Err(ConstructError::EinsumOutputNotSubset {
+                    circuit_name: out.name.clone(),
+                    all_input_axes: map.into_keys().collect(),
+                    output_axes: out.out_axes.clone(),
+                }
+                .into());
             }
         }
 
@@ -229,15 +252,7 @@ impl CircuitNodeAutoName for Einsum {
             } else {
                 Some(
                     self.input_circuits()
-                        .filter_map(|x| {
-                            x.name().map(|y| {
-                                if y.len() > 100 {
-                                    "...".to_owned()
-                                } else {
-                                    y.to_owned()
-                                }
-                            })
-                        })
+                        .map(|x| Self::shorten_child_name(x.name().unwrap()))
                         .collect::<Vec<String>>()
                         .join(" * "),
                 )
@@ -295,6 +310,23 @@ impl Einsum {
         self.input_axes().cloned().map(PyEinsumAxes).collect()
     }
 
+    #[staticmethod]
+    pub fn from_nodes_ints(
+        nodes: Vec<CircuitRc>,
+        input_ints: Vec<EinsumAxes>,
+        output_ints: EinsumAxes,
+        name: Option<String>,
+    ) -> Result<Self> {
+        if nodes.len() != input_ints.len() {
+            bail!(ConstructError::EinsumWrongNumChildren {
+                actual_num_children: nodes.len(),
+                expected_num_children_based_on_axes: input_ints.len(),
+                einsum_name: name
+            })
+        }
+        Self::try_new(zip(nodes, input_ints).collect(), output_ints, name)
+    }
+
     #[staticmethod]
     #[args(nodes = "*", name)]
     pub fn from_einsum_string(
@@ -302,8 +334,11 @@ impl Einsum {
         nodes: Vec<CircuitRc>,
         name: Option<String>,
     ) -> Result<Self> {
-        let (input_ints, output_ints) = EinsumSpec::string_to_ints(string.to_owned())?;
-        Self::try_new(zip(nodes, input_ints).collect(), output_ints, name)
+        let (mut input_ints, output_ints) = EinsumSpec::string_to_ints(string.to_owned())?;
+        if input_ints.len() == 1 && input_ints[0].is_empty() && nodes.len() == 0 {
+            input_ints.clear()
+        }
+        Self::from_nodes_ints(nodes, input_ints, output_ints, name)
     }
 
     #[staticmethod]
@@ -313,8 +348,11 @@ impl Einsum {
         nodes: Vec<CircuitRc>,
         name: Option<String>,
     ) -> Result<Self> {
-        let (input_ints, output_ints) = EinsumSpec::fancy_string_to_ints(string.to_owned())?;
-        Self::try_new(zip(nodes, input_ints).collect(), output_ints, name)
+        let (mut input_ints, output_ints) = EinsumSpec::fancy_string_to_ints(string.to_owned())?;
+        if input_ints.len() == 1 && input_ints[0].is_empty() && nodes.len() == 0 {
+            input_ints.clear()
+        }
+        Self::from_nodes_ints(nodes, input_ints, output_ints, name)
     }
 
     #[staticmethod]
@@ -350,7 +388,7 @@ impl Einsum {
         Einsum::try_new(
             vec![
                 (node, axes.clone()),
-                (ScalarConstant::new(scalar, sv![], scalar_name).rc(), sv![]),
+                (Scalar::new(scalar, sv![], scalar_name).rc(), sv![]),
             ],
             axes,
             name,
@@ -508,9 +546,9 @@ impl Add {
             name: Default::default(),
             info: Default::default(),
         };
-        out.name = out.auto_name(name);
+        out.name = out.auto_name(name.clone());
         out.compute_shape_maybe()
-            .context("Sum isn't broadcastable")?;
+            .with_context(|| format!("Sum isn't broadcastable name={:?}", name))?;
         out.init_info()
     }
 
@@ -591,23 +629,19 @@ impl Add {
     }
 }
 
-circuit_node_extra_impl!(Add);
+circuit_node_extra_impl!(Add, self_hash_default);
+
+impl CircuitNodeComputeInfoImpl for Add {
+    fn compute_shape(&self) -> Shape {
+        self.compute_shape_maybe().unwrap() // assuming error was caught on creation, panicking now
+    }
+}
+
+impl CircuitNodeHashItems for Add {}
 
 impl CircuitNode for Add {
     circuit_node_auto_impl!("88fb29e5-c81e-47fe-ae4c-678b22994670");
 
-    fn compute_shape(&self) -> Shape {
-        self.compute_shape_maybe().unwrap() // assuming error was caught on creation, panicking now
-    }
-
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
-        for node in &self.nodes {
-            hasher.update(&node.info().hash);
-        }
-
-        hasher
-    }
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         self.nodes_and_rank_differences()
             .iter()
@@ -672,15 +706,7 @@ impl CircuitNodeAutoName for Add {
             } else {
                 Some(
                     self.children()
-                        .filter_map(|x| {
-                            x.name().map(|y| {
-                                if y.len() > 100 {
-                                    "...".to_owned()
-                                } else {
-                                    y.to_owned()
-                                }
-                            })
-                        })
+                        .map(|x| Self::shorten_child_name(x.name().unwrap()))
                         .collect::<Vec<String>>()
                         .join(" + "),
                 )
@@ -711,20 +737,19 @@ impl Rearrange {
             name: Default::default(),
             info: Default::default(),
         };
-        out.name = out.auto_name(name);
+        out.name = out.auto_name(name.clone());
 
         // Check that RearrangeSpec is valid for input, prevent panic on rearrange.info().shape
-        if !out
-            .spec
+        out.spec
             .conform_to_input_shape(&node.info().shape, false)
-            .is_ok()
-        {
-            return Err(ConstructError::RearrangeWrongInputShape {
-                spec: out.spec,
-                shape: node.info().shape.clone(),
-            }
-            .into());
-        }
+            .with_context(|| {
+                format!(
+                    "{}\nname={:?}, child={:?}",
+                    "rearrange spec conform_to_input_shape failed in construct rearrange",
+                    name,
+                    node
+                )
+            })?;
 
         // check that known sizes of each axis divide the operand shape
         let input_sizes: Vec<u64> = out
@@ -790,21 +815,22 @@ impl Rearrange {
     }
 }
 
-circuit_node_extra_impl!(Rearrange);
-
-impl CircuitNode for Rearrange {
-    circuit_node_auto_impl!("13204d30-2f12-4edd-8765-34bc8b458ef2");
+circuit_node_extra_impl!(Rearrange, self_hash_default);
 
+impl CircuitNodeComputeInfoImpl for Rearrange {
     fn compute_shape(&self) -> Shape {
         self.conform_to_input_shape_spec(false).shapes().unwrap().1
     }
+}
 
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
-        hasher.update(&self.node.info().hash);
+impl CircuitNodeHashItems for Rearrange {
+    fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
         hasher.update(&self.spec.compute_hash());
-        hasher
     }
+}
+
+impl CircuitNode for Rearrange {
+    circuit_node_auto_impl!("13204d30-2f12-4edd-8765-34bc8b458ef2");
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         vec![self
@@ -989,21 +1015,33 @@ impl Index {
     }
 }
 
-circuit_node_extra_impl!(Index);
-
-impl CircuitNode for Index {
-    circuit_node_auto_impl!("3c655670-b352-4a5f-891c-0d7160609341");
+circuit_node_extra_impl!(Index, self_hash_default);
 
+impl CircuitNodeComputeInfoImpl for Index {
     fn compute_shape(&self) -> Shape {
         self.index.apply_to_shape(&self.node.info().shape)
     }
 
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
-        hasher.update(&self.node.info().hash);
-        hasher.update(&self.index.compute_hash());
-        hasher
+    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
+        Box::new(self.index.0.iter().filter_map(|x| match x {
+            TensorAxisIndex::Tensor(tensor) if !tensor.shape().is_empty() => {
+                let mut out: TorchDeviceDtypeOp = TorchDeviceDtype::from_tensor(tensor).into();
+                out.dtype = None;
+                Some(out)
+            }
+            _ => None,
+        }))
     }
+}
+
+impl CircuitNodeHashItems for Index {
+    fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
+        hasher.update(&self.index.compute_hash());
+    }
+}
+
+impl CircuitNode for Index {
+    circuit_node_auto_impl!("3c655670-b352-4a5f-891c-0d7160609341");
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         let mut cur: i32 = -1;
@@ -1067,17 +1105,6 @@ impl CircuitNode for Index {
             0
         }
     }
-
-    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
-        Box::new(self.index.0.iter().filter_map(|x| match x {
-            TensorAxisIndex::Tensor(tensor) if !tensor.shape().is_empty() => {
-                let mut out: TorchDeviceDtypeOp = TorchDeviceDtype::from_tensor(tensor).into();
-                out.dtype = None;
-                Some(out)
-            }
-            _ => None,
-        }))
-    }
 }
 
 impl CircuitNodeAutoName for Index {
@@ -1268,6 +1295,12 @@ impl Concat {
         }
         let node_shape = &nodes[0].info().shape;
         for node in nodes.iter() {
+            if node.ndim() != nodes[0].ndim() {
+                return Err(ConstructError::ConcatShapeDifferent {
+                    shapes: nodes.iter().map(|x| x.info().shape.clone()).collect(),
+                }
+                .into());
+            }
             for (i, s) in node.info().shape.iter().enumerate() {
                 if i != axis && *s != node_shape[i] {
                     return Err(ConstructError::ConcatShapeDifferent {
@@ -1297,25 +1330,24 @@ impl Concat {
     }
 }
 
-circuit_node_extra_impl!(Concat);
-
-impl CircuitNode for Concat {
-    circuit_node_auto_impl!("f2684583-c215-4f67-825e-6e4e51091ca7");
+circuit_node_extra_impl!(Concat, self_hash_default);
 
+impl CircuitNodeComputeInfoImpl for Concat {
     fn compute_shape(&self) -> Shape {
         let mut shape = self.nodes[0].info().shape.clone();
         shape[self.axis] = self.nodes.iter().map(|n| n.info().shape[self.axis]).sum();
         shape
     }
+}
 
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
-        for node in &self.nodes {
-            hasher.update(&node.info().hash);
-        }
+impl CircuitNodeHashItems for Concat {
+    fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
         hasher.update(&self.axis.to_le_bytes());
-        hasher
     }
+}
+
+impl CircuitNode for Concat {
+    circuit_node_auto_impl!("f2684583-c215-4f67-825e-6e4e51091ca7");
 
     fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(self.nodes.iter().cloned())
@@ -1369,15 +1401,7 @@ impl CircuitNodeAutoName for Concat {
             } else {
                 Some(
                     self.children()
-                        .filter_map(|x| {
-                            x.name().map(|y| {
-                                if y.len() > 100 {
-                                    "...".to_owned()
-                                } else {
-                                    y.to_owned()
-                                }
-                            })
-                        })
+                        .map(|x| Self::shorten_child_name(x.name().unwrap()))
                         .collect::<Vec<String>>()
                         .join(" ++ "),
                 )
@@ -1487,24 +1511,25 @@ impl Scatter {
     }
 }
 
-circuit_node_extra_impl!(Scatter);
-
-impl CircuitNode for Scatter {
-    circuit_node_auto_impl!("50cce6d3-457c-4f52-9a23-1903bdc76533");
+circuit_node_extra_impl!(Scatter, self_hash_default);
 
+impl CircuitNodeComputeInfoImpl for Scatter {
     fn compute_shape(&self) -> Shape {
         self.info().shape.clone()
     }
+}
 
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
-        hasher.update(&self.node.info().hash);
+impl CircuitNodeHashItems for Scatter {
+    fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
         hasher.update(&self.index.compute_hash());
         for i in &self.info().shape {
             hasher.update(&i.to_le_bytes());
         }
-        hasher
     }
+}
+
+impl CircuitNode for Scatter {
+    circuit_node_auto_impl!("50cce6d3-457c-4f52-9a23-1903bdc76533");
 
     fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(std::iter::once(self.node.clone()))
@@ -1762,26 +1787,26 @@ impl Conv {
     }
 }
 
-circuit_node_extra_impl!(Conv);
-
-impl CircuitNode for Conv {
-    circuit_node_auto_impl!("a329465f-a18f-4313-a2a6-51d8ebad7dd6");
+circuit_node_extra_impl!(Conv, self_hash_default);
 
+impl CircuitNodeComputeInfoImpl for Conv {
     fn compute_shape(&self) -> Shape {
         self.info().shape.clone()
     }
-
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
-        hasher.update(&self.input.info().hash);
-        hasher.update(&self.filter.info().hash);
+}
+impl CircuitNodeHashItems for Conv {
+    fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
+        // this is ok because all looped over items are fixed len and we follow this with uuid (from children hash)
         for (i, (j, k)) in zip(&self.stride, &self.padding) {
             hasher.update(&i.to_le_bytes());
             hasher.update(&j.to_le_bytes());
             hasher.update(&k.to_le_bytes());
         }
-        hasher
     }
+}
+
+impl CircuitNode for Conv {
+    circuit_node_auto_impl!("a329465f-a18f-4313-a2a6-51d8ebad7dd6");
 
     fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new([self.input.clone(), self.filter.clone()].into_iter())
