@@ -17,10 +17,22 @@ use rustc_hash::FxHashMap as HashMap;
 use uuid::Uuid;
 
 use crate::{
-    circuit_node_auto_impl, circuit_node_extra_impl, named_axes::set_named_axes, new_rc,
-    prelude::*, CachedCircuitInfo, NamedAxes, PyCircuitBase, Shape, TensorEvalError,
+    circuit_node_auto_impl, circuit_node_extra_impl,
+    circuit_node_private::{CircuitNodeComputeInfoImpl, CircuitNodeHashItems},
+    named_axes::set_named_axes,
+    new_rc,
+    prelude::*,
+    CachedCircuitInfo, NamedAxes, PyCircuitBase, Shape, TensorEvalError,
 };
 
+macro_rules! compute_info_auto_leaf_impl {
+    () => {
+        fn compute_max_non_leaf_size(&self) -> BigUint {
+            BigUint::from(0usize)
+        }
+    };
+}
+
 macro_rules! circuit_node_auto_leaf_impl {
     ($uuid:literal) => {
         circuit_node_auto_impl!($uuid);
@@ -44,32 +56,35 @@ macro_rules! circuit_node_auto_leaf_impl {
 
 #[pyclass(extends=PyCircuitBase, unsendable)]
 #[derive(Clone)]
-pub struct ArrayConstant {
+pub struct Array {
     #[pyo3(get)]
     pub value: Tensor,
     info: CachedCircuitInfo,
     name: Option<String>,
 }
 
-circuit_node_extra_impl!(ArrayConstant);
-
-impl CircuitNode for ArrayConstant {
-    circuit_node_auto_leaf_impl!("b2aac9d5-1bfa-4c2a-9684-e3f9ecbc1b94");
+circuit_node_extra_impl!(Array, self_hash_default);
 
+impl CircuitNodeComputeInfoImpl for Array {
     fn compute_shape(&self) -> Shape {
         self.value.shape().clone()
     }
+    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
+        Box::new(std::iter::once(
+            TorchDeviceDtype::from_tensor(&self.value).into(),
+        ))
+    }
+    compute_info_auto_leaf_impl!();
+}
 
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
+impl CircuitNodeHashItems for Array {
+    fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
         hasher.update(self.value.hash().unwrap());
-
-        hasher
     }
+}
 
-    fn max_non_input_size(&self) -> BigUint {
-        Default::default()
-    }
+impl CircuitNode for Array {
+    circuit_node_auto_leaf_impl!("b2aac9d5-1bfa-4c2a-9684-e3f9ecbc1b94");
 
     fn eval_tensors(
         &self,
@@ -85,15 +100,9 @@ impl CircuitNode for ArrayConstant {
         // for them.
         0
     }
-
-    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
-        Box::new(std::iter::once(
-            TorchDeviceDtype::from_tensor(&self.value).into(),
-        ))
-    }
 }
 
-impl ArrayConstant {
+impl Array {
     #[apply(new_rc)]
     pub fn new(value: Tensor, name: Option<String>) -> (Self) {
         let value = value.hashed();
@@ -113,9 +122,9 @@ impl ArrayConstant {
         if let Some(tc) = tensor_cache {
             return tc
                 .get_tensor(hash_base16.to_owned())
-                .map(|value| ArrayConstant::new(value, name));
+                .map(|value| Array::new(value, name));
         }
-        get_tensor_prefix(hash_base16).map(|value| ArrayConstant::new(value, name))
+        get_tensor_prefix(hash_base16).map(|value| Array::new(value, name))
     }
 
     pub fn randn(shape: Shape) -> Self {
@@ -132,10 +141,10 @@ impl ArrayConstant {
 }
 
 #[pymethods]
-impl ArrayConstant {
+impl Array {
     #[new]
     fn py_new(value: Tensor, name: Option<String>) -> PyClassInitializer<Self> {
-        ArrayConstant::new(value, name).into_init()
+        Array::new(value, name).into_init()
     }
 
     #[staticmethod]
@@ -176,7 +185,7 @@ impl ArrayConstant {
             if let Some(device) = device_dtype.device {
                 kwargs.insert("device", device.into_py(py));
             }
-            ArrayConstant::new(
+            Array::new(
                 PY_UTILS
                     .torch
                     .getattr(py, "randn")
@@ -204,7 +213,7 @@ impl ArrayConstant {
 
     #[staticmethod]
     pub fn from_hash(name: Option<String>, hash_base16: &str) -> Result<Self> {
-        get_tensor_prefix(hash_base16).map(|value| ArrayConstant::new(value, name))
+        get_tensor_prefix(hash_base16).map(|value| Array::new(value, name))
     }
 
     #[staticmethod]
@@ -215,7 +224,7 @@ impl ArrayConstant {
         tensor_cache: Option<TensorCacheRrfs>,
     ) -> Result<Self> {
         let mut tensor_cache = tensor_cache;
-        ArrayConstant::from_hash_prefix(name, hash_base16, &mut tensor_cache)
+        Array::from_hash_prefix(name, hash_base16, &mut tensor_cache)
     }
 }
 
@@ -228,32 +237,36 @@ pub struct Symbol {
     name: Option<String>,
 }
 
-circuit_node_extra_impl!(Symbol);
-
-impl CircuitNode for Symbol {
-    circuit_node_auto_leaf_impl!("13c3ee63-76e9-4afb-8057-40309d17b458");
+circuit_node_extra_impl!(Symbol, self_hash_default);
 
+impl CircuitNodeComputeInfoImpl for Symbol {
     fn compute_shape(&self) -> Shape {
         self.info.shape.clone() // note: assumes shape has already been initted!
     }
-
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
-        hasher.update(self.uuid.as_bytes());
-
-        hasher
-    }
-
     fn compute_is_explicitly_computable(&self) -> bool {
         false
     }
+    compute_info_auto_leaf_impl!();
+}
+
+impl CircuitNodeHashItems for Symbol {
+    fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
+        hasher.update(self.uuid.as_bytes());
+        for l in &self.info.shape {
+            hasher.update(&l.to_le_bytes());
+        }
+    }
+}
+
+impl CircuitNode for Symbol {
+    circuit_node_auto_leaf_impl!("13c3ee63-76e9-4afb-8057-40309d17b458");
 
     fn eval_tensors(
         &self,
         _tensors: &[Tensor],
         _device_dtype: &TorchDeviceDtype,
     ) -> Result<Tensor> {
-        Err(TensorEvalError::NotExplicitlyComputable {
+        Err(TensorEvalError::NotExplicitlyComputableInternal {
             circuit: self.crc(),
         }
         .into())
@@ -296,30 +309,33 @@ impl Symbol {
 
 #[pyclass(extends=PyCircuitBase, unsendable)]
 #[derive(Clone)]
-pub struct ScalarConstant {
+pub struct Scalar {
     #[pyo3(get)]
     pub value: f64,
     info: CachedCircuitInfo,
     name: Option<String>,
 }
 
-circuit_node_extra_impl!(ScalarConstant);
-
-impl CircuitNode for ScalarConstant {
-    circuit_node_auto_leaf_impl!("78a77905-8b3f-4471-bb77-255673941fef");
+circuit_node_extra_impl!(Scalar, self_hash_default);
 
+impl CircuitNodeComputeInfoImpl for Scalar {
     fn compute_shape(&self) -> Shape {
         self.info().shape.clone() // note: assumes shape has already been initted!
     }
+    compute_info_auto_leaf_impl!();
+}
 
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
+impl CircuitNodeHashItems for Scalar {
+    fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
         hasher.update(&self.value.to_le_bytes());
         for l in &self.info.shape {
             hasher.update(&l.to_le_bytes());
         }
-        hasher
     }
+}
+
+impl CircuitNode for Scalar {
+    circuit_node_auto_leaf_impl!("78a77905-8b3f-4471-bb77-255673941fef");
 
     fn eval_tensors(&self, _tensors: &[Tensor], device_dtype: &TorchDeviceDtype) -> Result<Tensor> {
         Ok(scalar_to_tensor(
@@ -335,7 +351,7 @@ impl CircuitNode for ScalarConstant {
     }
 }
 
-impl ScalarConstant {
+impl Scalar {
     #[apply(new_rc)]
     pub fn new(value: f64, shape: Shape, name: Option<String>) -> (Self) {
         let mut out = Self {
@@ -350,7 +366,7 @@ impl ScalarConstant {
 }
 
 #[pymethods]
-impl ScalarConstant {
+impl Scalar {
     #[new]
     #[args(shape = "sv![]")]
     fn py_new(value: f64, shape: Shape, name: Option<String>) -> PyClassInitializer<Self> {
@@ -373,6 +389,6 @@ impl ScalarConstant {
 #[test]
 fn test_nrc() {
     pyo3::prepare_freethreaded_python();
-    let ex = ScalarConstant::nrc(0.0, sv![1, 2], None);
+    let ex = Scalar::nrc(0.0, sv![1, 2], None);
     ex.print().unwrap();
 }
