@@ -21,8 +21,11 @@ use thiserror::Error;
 use uuid::uuid;
 
 use crate::{
-    circuit_node_auto_impl, circuit_node_extra_impl, new_rc_unwrap, prelude::*, ArrayConstant,
-    CachedCircuitInfo, HashBytes, PyCircuitBase,
+    circuit_node_auto_impl, circuit_node_extra_impl,
+    circuit_node_private::{CircuitNodeComputeInfoImpl, CircuitNodeHashItems},
+    new_rc_unwrap,
+    prelude::*,
+    Array, CachedCircuitInfo, HashBytes, PyCircuitBase,
 };
 
 macro_rules! gf_gen {
@@ -71,6 +74,7 @@ gf_gen!(
     // (q_from_qr, 2), // TODO: this requires first dim > second dim!
     (min, 0, 1),
     (max, 0, 1),
+    (last_dim_size, 0, 1),
 );
 
 static SPECS: GILLazy<HashMap<String, GeneralFunctionSpec>> = GILLazy::new(|| {
@@ -602,23 +606,22 @@ impl GeneralFunction {
     }
 }
 
-circuit_node_extra_impl!(GeneralFunction);
-
-impl CircuitNode for GeneralFunction {
-    circuit_node_auto_impl!("3c655670-b352-4a5f-891c-0d7160609341");
+circuit_node_extra_impl!(GeneralFunction, self_hash_default);
 
+impl CircuitNodeComputeInfoImpl for GeneralFunction {
     fn compute_shape(&self) -> Shape {
         self.info.shape.clone() // set in try new!
     }
+}
 
-    fn compute_hash(&self) -> blake3::Hasher {
-        let mut hasher = blake3::Hasher::new();
+impl CircuitNodeHashItems for GeneralFunction {
+    fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
         hasher.update(&self.spec.compute_hash());
-        for node in &self.nodes {
-            hasher.update(&node.info().hash);
-        }
-        hasher
     }
+}
+
+impl CircuitNode for GeneralFunction {
+    circuit_node_auto_impl!("3c655670-b352-4a5f-891c-0d7160609341");
 
     fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(self.nodes.iter().cloned())
@@ -673,17 +676,9 @@ impl CircuitNodeAutoName for GeneralFunction {
                         + " "
                         + &self
                             .children()
-                            .filter_map(|x| {
-                                x.name().map(|y| {
-                                    if y.len() > 100 {
-                                        "...".to_owned()
-                                    } else {
-                                        y.to_owned()
-                                    }
-                                })
-                            })
+                            .map(|x| Self::shorten_child_name(x.name().unwrap()))
                             .collect::<Vec<String>>()
-                            .join(" , "),
+                            .join(", "),
                 )
             }
         })
@@ -1043,7 +1038,7 @@ impl GeneralFunctionSpecTester {
                 // TODO: swap to f64 as needed!
                 let tensors: Vec<_> = shapes
                     .iter()
-                    .map(|shape| ArrayConstant::randn(shape.clone()).value)
+                    .map(|shape| Array::randn(shape.clone()).value)
                     .collect();
 
                 let out_tensor = spec
