@@ -2,7 +2,7 @@
 #![allow(clippy::borrow_deref_ref)]
 use std::{
     collections::BTreeMap,
-    fmt::Debug,
+    fmt::{self, Debug},
     hash::Hash,
     iter::zip,
     ops::{Deref, DerefMut},
@@ -14,12 +14,12 @@ use circuit_utils::toposort_circuit;
 pub use computational_node::{
     flat_concat, flat_concat_back, Add, Concat, Conv, Einsum, Index, Rearrange, Scatter,
 };
-pub use constant::{ArrayConstant, ScalarConstant, Symbol};
+pub use constant::{Array, Scalar, Symbol};
 pub use cumulant::Cumulant;
 pub use generalfunction::{GeneralFunction, GeneralFunctionSpec};
 use itertools::Itertools;
 use macro_rules_attribute::apply;
-pub use module_nodes::{ModuleNode, ModuleNodeArgSpec, ModuleNodeSpec};
+pub use module::{Module, ModuleArgSpec, ModuleSpec};
 use num_bigint::BigUint;
 use print::PrintOptions;
 use pyo3::{
@@ -35,102 +35,38 @@ use rr_util::{
     pycall, pycallable, python_error_exception,
     rearrange_spec::{OpSize, PyCountsAtAxes, RInnerInts, RearrangeSpec},
     sv,
+    symbolic_size::{SymbolicSizeConstraints, SymbolicSizeProduct},
     tensor_util::{check_canon_idxs, Shape, TensorIndex, TorchDeviceDtype, TorchDeviceDtypeOp},
-    util::{arc_ref_clone, arc_unwrap_or_clone, AsOp, AxisInt, HashBytes, HashBytesToPy},
+    util::{
+        arc_ref_clone, arc_unwrap_or_clone, AsOp, AxisInt, EinsumAxes, HashBytes, HashBytesToPy,
+    },
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
+pub use set_symbolic_size::SetSymbolicShape;
 use thiserror::Error;
-use uuid::uuid;
 pub use variable_nodes::{DiscreteVar, StoredCumulantVar, Tag};
 
-use crate::print::print_circuit_stats;
+use crate::{module::inline_all_modules, print::print_circuit_stats};
 pub mod circuit_utils;
 pub mod computational_node;
 pub mod constant;
 pub mod cumulant;
 pub mod expand_node;
 pub mod generalfunction;
-pub mod module_nodes;
+pub mod module;
 pub mod named_axes;
 pub mod nrc;
 pub mod opaque_iterative_matcher;
 pub mod parsing;
 pub mod print;
+pub mod set_symbolic_size;
 pub mod variable_nodes;
 
-mod circuit_node_private {
-    use crate::{CachedCircuitInfo, Result};
-    pub trait CircuitNodePrivate {
-        fn info_mut(&mut self) -> &mut CachedCircuitInfo;
-        fn name_mut(&mut self) -> &mut Option<String>;
-    }
-
-    pub trait CircuitNodeInit {
-        fn init_info_impl(self) -> Result<Self>
-        where
-            Self: Sized;
-
-        fn rename_impl(self, new_name: Option<String>) -> Self
-        where
-            Self: Sized;
-
-        fn update_info_impl<F>(self, f: F) -> Result<Self>
-        where
-            Self: Sized,
-            F: FnOnce(&mut CachedCircuitInfo);
-    }
-}
+mod circuit_node_private;
 use circuit_node_private::*;
 
-use crate::circuit_utils::total_arrayconstant_size;
-
-impl<T: CircuitNodePrivate + CircuitNode> CircuitNodeInit for T {
-    fn init_info_impl(mut self) -> Result<Self>
-    where
-        Self: Sized,
-    {
-        self.info_mut().shape = self.compute_shape(); // set shape so methods to compute other info can use it
-
-        let mut hasher = self.compute_hash();
-        self.info_mut().named_axes = self.compute_named_axes();
-        hasher.update(self.name().unwrap_or("").as_bytes());
-        hasher.update(&self.node_type_uuid());
-        hasher.update(uuid!("025e9af4-1366-4211-aa5f-7c28fc6cdf9f").as_bytes());
-        for (axis, name) in &self.info().named_axes {
-            if *axis as usize >= self.info().shape.len() {
-                return Err(ConstructError::NamedAxisAboveRank {}.into());
-            }
-            hasher.update(&[*axis]);
-            hasher.update(name.as_bytes());
-        }
-        self.info_mut().is_constant = self.compute_is_constant();
-        self.info_mut().is_explicitly_computable = self.compute_is_explicitly_computable();
-        self.info_mut().can_be_sampled = self.compute_can_be_sampled();
-        self.info_mut().hash = hasher.finalize().into();
-        self.info_mut().max_non_input_size = self.max_non_input_size();
-        self.info_mut().device_dtype = self.compute_device_dtype()?;
-        Ok(self)
-    }
-
-    fn rename_impl(mut self, new_name: Option<String>) -> Self
-    where
-        Self: Sized,
-    {
-        *self.name_mut() = new_name;
-        self.init_info_impl().unwrap() // we could avoid recomputing some stuff if we wanted
-    }
-
-    fn update_info_impl<F>(mut self, f: F) -> Result<Self>
-    where
-        Self: Sized,
-        F: FnOnce(&mut CachedCircuitInfo),
-    {
-        f(self.info_mut());
-        self.init_info_impl()
-    }
-}
-
 use self::circuit_utils::total_flops;
+use crate::circuit_utils::total_arrayconstant_size;
 
 fn check_canon_axes(rank: usize, axes: &[i64]) -> Result<Vec<usize>> {
     check_canon_idxs(rank, axes).context("axis out of bounds")
@@ -143,43 +79,56 @@ fn check_canon_axes_as(rank: usize, axes: &[i64]) -> Result<RInnerInts> {
 }
 
 pub type NamedAxes = BTreeMap<AxisInt, String>;
-pub trait CircuitNode: CircuitNodeInit {
+
+pub trait CircuitNodeSelfOnlyHash {
+    fn compute_self_only_hash(&self, hasher: &mut blake3::Hasher);
+}
+
+#[macro_export]
+macro_rules! circuit_node_self_only_hash_default_impl {
+    ($type_name:ty) => {
+        impl $crate::CircuitNodeSelfOnlyHash for $type_name {
+            fn compute_self_only_hash(&self, hasher: &mut blake3::Hasher) {
+                use $crate::circuit_node_private::CircuitNodeHashItems;
+                self.compute_hash_non_name_non_children(hasher)
+            }
+        }
+    };
+}
+
+pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     // ==== implementable section ===
     //
     // NOTE: ALL FNS IN THIS SECTION *MUST* BE COPIED TO THE CIRCUIT NODE UNION IMPL!
     // If you add something here with a default impl, write a new impl for circuit node union!
-    // (up until default section)
+    // (up until default only section)
     //
     // we could enforce this sort of stuff with some proc macros, but seems like overkill atm.
 
     fn info(&self) -> &CachedCircuitInfo;
     fn name(&self) -> Option<&str>;
 
-    fn compute_shape(&self) -> Shape;
-    fn compute_hash(&self) -> blake3::Hasher; // shouldn't hash name
-    fn compute_is_constant(&self) -> bool {
-        self.children().all(|c| c.info().is_constant)
-    }
-    fn compute_is_explicitly_computable(&self) -> bool {
-        self.children().all(|c| c.info().is_explicitly_computable)
-    }
-    fn compute_can_be_sampled(&self) -> bool {
-        self.children().all(|c| c.info().can_be_sampled)
-    }
-
-    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
-        Box::new(std::iter::empty())
-    }
-
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>>;
 
     fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_>;
 
+    fn non_free_children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
+        self.children()
+    }
+
     fn map_children_enumerate<F>(&self, f: F) -> Result<Self>
     where
         Self: Sized,
         F: FnMut(usize, CircuitRc) -> Result<CircuitRc>;
 
+    fn map_non_free_children_enumerate<F>(&self, f: F) -> Result<Self>
+    where
+        Self: Sized,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc>,
+    {
+        self.map_children_enumerate(f)
+    }
+
     fn node_type_uuid(&self) -> [u8; 16];
 
     fn self_flops(&self) -> BigUint {
@@ -196,6 +145,19 @@ pub trait CircuitNode: CircuitNodeInit {
 
     fn c(self) -> Circuit;
     fn rc(self) -> CircuitRc;
+
+    /// for nodes with fixed uuid, should be same as node_type_uuid
+    fn static_node_type_uuid() -> [u8; 16]
+    where
+        Self: Sized,
+    {
+        *uuid::Uuid::nil().as_bytes()
+    }
+
+    // ==== default only section ===
+    // FUNCTIONS BELOW HERE *shouldn't* be overridden by implementors!
+    // (if you do implement, this won't be picked up on by CircuitNodeDefer (union types etc)!)
+
     fn init_info(self) -> Result<Self>
     where
         Self: Sized,
@@ -216,9 +178,10 @@ pub trait CircuitNode: CircuitNodeInit {
         self.update_info_impl(f)
     }
 
-    // ==== default section ===
-    // FUNCTIONS BELOW HERE *shouldn't* be overridden by implementors!
-    // (if you do implement, this won't be picked up on by union types!)
+    // implementation helper
+    fn already_set_named_axes(&self) -> Option<NamedAxes> {
+        (!self.info().named_axes.is_empty()).then(|| self.info().named_axes.clone())
+    }
 
     fn crc(&self) -> CircuitRc
     where
@@ -230,33 +193,7 @@ pub trait CircuitNode: CircuitNodeInit {
         self.name().map(|x| x.to_owned())
     }
 
-    fn compute_device_dtype(&self) -> Result<TorchDeviceDtypeOp> {
-        self.children()
-            .map(|c| c.info().device_dtype.clone())
-            .chain(self.device_dtype_extra())
-            .fold(Ok(TorchDeviceDtypeOp::NONE), |acc, new| {
-                acc.map(|old| TorchDeviceDtypeOp::combine(old, new))?
-            })
-    }
-
-    fn compute_named_axes(&self) -> NamedAxes {
-        if !self.info().named_axes.is_empty() {
-            return self.info().named_axes.clone();
-        }
-        if !self.children().any(|x| !x.info().named_axes.is_empty()) {
-            return BTreeMap::new();
-        }
-        let child_axis_map = self.child_axis_map();
-        let mut result: NamedAxes = BTreeMap::new();
-        for (mp, child) in zip(child_axis_map, self.children()) {
-            for (ax, name) in &child.info().named_axes {
-                if let Some(top_ax) = mp[(*ax) as usize] {
-                    result.insert(top_ax as u8, name.clone());
-                }
-            }
-        }
-        result
-    }
+    // TODO: add more map utils for non_free as needed
 
     fn map_children<F>(&self, mut f: F) -> Result<Self>
     where
@@ -265,6 +202,13 @@ pub trait CircuitNode: CircuitNodeInit {
     {
         self.map_children_enumerate(|_i, x| f(x))
     }
+    fn map_non_free_children<F>(&self, mut f: F) -> Result<Self>
+    where
+        Self: Sized,
+        F: FnMut(CircuitRc) -> Result<CircuitRc>,
+    {
+        self.map_non_free_children_enumerate(|_i, x| f(x))
+    }
 
     fn map_children_idxs<F>(&self, mut f: F) -> Result<Self>
     where
@@ -281,6 +225,13 @@ pub trait CircuitNode: CircuitNodeInit {
     {
         self.map_children(|x| Ok(f(x))).unwrap()
     }
+    fn map_non_free_children_unwrap<F>(&self, mut f: F) -> Self
+    where
+        Self: Sized,
+        F: FnMut(CircuitRc) -> CircuitRc,
+    {
+        self.map_non_free_children(|x| Ok(f(x))).unwrap()
+    }
 
     fn map_children_unwrap_enumerate<F>(&self, mut f: F) -> Self
     where
@@ -320,19 +271,11 @@ pub trait CircuitNode: CircuitNodeInit {
         }
     }
 
-    fn max_non_input_size(&self) -> BigUint {
-        self.children()
-            .map(|x| x.info().max_non_input_size.clone())
-            .chain(std::iter::once(self.info().numel()))
-            .max()
-            .unwrap_or(0usize.into())
-    }
-
     fn repr(&self) -> Result<String>
     where
         Self: Clone,
     {
-        PrintOptions::repr_deep(&Default::default(), self.crc())
+        PrintOptions::repr(&Default::default(), self.crc())
     }
 
     fn print(&self) -> Result<()>
@@ -410,10 +353,8 @@ pub trait CircuitNode: CircuitNodeInit {
             .into_iter()
             .map(|i| self.info().shape[i as usize])
             .product();
-        self.sum(axes, name)?.mul(
-            ScalarConstant::nrc(1. / (total_size as f64), sv![], None),
-            None,
-        )
+        self.sum(axes, name)?
+            .mul(Scalar::nrc(1. / (total_size as f64), sv![], None), None)
     }
 
     fn reduce(&self, op_name: String, axes: &[i64], name: Option<String>) -> Result<Circuit>
@@ -483,7 +424,7 @@ pub trait CircuitNode: CircuitNodeInit {
     where
         Self: Clone,
     {
-        self.mul(ScalarConstant::nrc(scalar, sv![], scalar_name), name)
+        self.mul(Scalar::nrc(scalar, sv![], scalar_name), name)
     }
     fn index(&self, index: TensorIndex, name: Option<String>) -> Result<Index>
     where
@@ -621,11 +562,24 @@ pub trait CircuitNode: CircuitNodeInit {
 
 pub trait CircuitNodeAutoName: CircuitNode {
     fn auto_name(&self, name: Option<String>) -> Option<String>;
+    const CHILD_NAME_MAX_LEN: usize = 100; // TODO: fix me!
+    /// we shorten children names on multi child nodes
+    fn shorten_child_name(name: &str) -> String {
+        if name.len() > Self::CHILD_NAME_MAX_LEN {
+            name[..Self::CHILD_NAME_MAX_LEN].to_owned() + "..."
+        } else {
+            name.to_owned()
+        }
+    }
 }
 
 pub trait CircuitNodeDefer: CircuitNodeInit {
     fn as_trait_obj(&self) -> &dyn CircuitNode;
     fn map_children_enumerate_impl<F>(&self, f: F) -> Result<Self>
+    where
+        Self: Sized,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc>;
+    fn map_non_free_children_enumerate_impl<F>(&self, f: F) -> Result<Self>
     where
         Self: Sized,
         F: FnMut(usize, CircuitRc) -> Result<CircuitRc>;
@@ -709,7 +663,7 @@ macro_rules! define_circuit_union_impl {
             }
         }
 
-        impl rr_util::eq_by_big_hash::EqByBigHash for $name{
+        impl rr_util::eq_by_big_hash::EqByBigHash for $name {
             fn hash(&self) -> rr_util::util::HashBytes {
                 self.info().hash
             }
@@ -758,6 +712,32 @@ macro_rules! define_circuit_union_impl {
                 }
             }
 
+            impl std::fmt::Display for [<$name Type>] {
+                fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+                    let out = match self {
+                        $(
+                        Self::$x => stringify!($x),
+                        )*
+                    };
+                    f.write_str(out)
+                }
+            }
+
+            impl std::str::FromStr for [<$name Type>] {
+                type Err = anyhow::Error;
+                fn from_str(s: &str) -> Result<Self, Self::Err> {
+                    let out = match s {
+                        $(
+                        stringify!($x) => Self::$x,
+                        )*
+                        v => bail!(rr_util::tensor_util::ParseError::InvalidVariant {
+                            v: v.to_owned()
+                        }),
+                    };
+                    Ok(out)
+                }
+            }
+
         }
 
 
@@ -852,6 +832,14 @@ macro_rules! define_circuit_union_impl {
                     (node, construct) => $crate::CircuitNode::map_children_enumerate(node, f).map(|v| construct(v)))
             }
 
+            fn map_non_free_children_enumerate_impl<F>(&self, f: F) -> Result<Self>
+            where
+                F: FnMut(usize,$crate::CircuitRc) -> Result<$crate::CircuitRc>,
+            {
+                [<$name:snake _map_construct>]!(self;
+                    (node, construct) => $crate::CircuitNode::map_non_free_children_enumerate(node, f).map(|v| construct(v)))
+            }
+
             fn custom_c(self) -> $crate::Circuit {
                 [<$name:snake _map>]!(self; node => node.c())
             }
@@ -867,11 +855,7 @@ macro_rules! define_circuit_union_impl {
                 type TypeTag = [<$name Type>];
 
                 fn variant_string(&self) -> String {
-                    match self {
-                        $(
-                            Self::$x(_) => stringify!($x).to_owned(),
-                        )*
-                    }
+                    self.type_tag().to_string()
                 }
 
                 fn type_tag(&self) -> Self::TypeTag {
@@ -929,9 +913,9 @@ macro_rules! define_circuit {
 
 define_circuit!(
     Einsum,
-    ArrayConstant,
+    Array,
     Symbol,
-    ScalarConstant,
+    Scalar,
     Add,
     Rearrange,
     Index,
@@ -939,7 +923,8 @@ define_circuit!(
     Concat,
     Scatter,
     Conv,
-    ModuleNode,
+    Module,
+    SetSymbolicShape,
     Tag,
     DiscreteVar,
     StoredCumulantVar,
@@ -1018,21 +1003,15 @@ macro_rules! define_circuit_union_special {
 // These nodes are uneffected by rewrites, and satisfy
 // AlgebraicRewrite(Replace(X, IrreducibleNode->Y)) == Replace(AlgebraicRewrite(IrreducibleNode), IrreducibleNode->Y)
 // except for hashmap iteration order or other unfortunate nondeterminism
-define_circuit_union_special!(IrreducibleNode {
-    ArrayConstant,
-    Symbol,
-});
+define_circuit_union_special!(IrreducibleNode { Array, Symbol });
 
 define_circuit_union_special!(Leaf {
-    ArrayConstant,
+    Array,
     Symbol,
-    ScalarConstant,
+    Scalar,
 });
 
-define_circuit_union_special!(LeafConstant {
-    ArrayConstant,
-    ScalarConstant,
-});
+define_circuit_union_special!(LeafConstant { Array, Scalar });
 
 define_circuit_union_special!(Var {
     StoredCumulantVar,
@@ -1042,6 +1021,10 @@ define_circuit_union_special!(Var {
 // work around for fact that we can't implement foreign trait on constrained type
 #[macro_export]
 macro_rules! circuit_node_extra_impl {
+    ($type_name:ident, self_hash_default) => {
+        $crate::circuit_node_extra_impl!($type_name);
+        $crate::circuit_node_self_only_hash_default_impl!($type_name);
+    };
     ($type_name:ident) => {
         $crate::circuit_node_eq_ord_debug!($type_name);
 
@@ -1087,6 +1070,12 @@ macro_rules! circuit_node_auto_impl {
             self.name.as_deref()
         }
         fn node_type_uuid(&self) -> [u8; 16] {
+            Self::static_node_type_uuid()
+        }
+        fn static_node_type_uuid() -> [u8; 16]
+        where
+            Self: Sized,
+        {
             *uuid::uuid!($the_uuid).as_bytes()
         }
         fn c(self) -> $crate::Circuit {
@@ -1098,6 +1087,12 @@ macro_rules! circuit_node_auto_impl {
     };
 }
 
+impl<T: CircuitNodeDefer> CircuitNodeSelfOnlyHash for T {
+    fn compute_self_only_hash(&self, hasher: &mut blake3::Hasher) {
+        self.as_trait_obj().compute_self_only_hash(hasher)
+    }
+}
+
 // UPDATE ME WHEN YOU CHANGE CircuitNode Trait!!!
 impl<T: CircuitNodeDefer> CircuitNode for T {
     fn info(&self) -> &CachedCircuitInfo {
@@ -1108,31 +1103,6 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.as_trait_obj().name()
     }
 
-    fn compute_shape(&self) -> Shape {
-        self.as_trait_obj().compute_shape()
-    }
-
-    fn compute_hash(&self) -> blake3::Hasher {
-        let hasher = self.as_trait_obj().compute_hash();
-        hasher
-    }
-
-    fn compute_is_constant(&self) -> bool {
-        self.as_trait_obj().compute_is_constant()
-    }
-
-    fn compute_is_explicitly_computable(&self) -> bool {
-        self.as_trait_obj().compute_is_explicitly_computable()
-    }
-
-    fn compute_can_be_sampled(&self) -> bool {
-        self.as_trait_obj().compute_can_be_sampled()
-    }
-
-    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
-        self.as_trait_obj().device_dtype_extra()
-    }
-
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         self.as_trait_obj().child_axis_map()
     }
@@ -1141,6 +1111,10 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.as_trait_obj().children()
     }
 
+    fn non_free_children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
+        self.as_trait_obj().non_free_children()
+    }
+
     fn map_children_enumerate<F>(&self, f: F) -> Result<Self>
     where
         F: FnMut(usize, CircuitRc) -> Result<CircuitRc>,
@@ -1148,6 +1122,13 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.map_children_enumerate_impl(f)
     }
 
+    fn map_non_free_children_enumerate<F>(&self, f: F) -> Result<Self>
+    where
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc>,
+    {
+        self.map_non_free_children_enumerate_impl(f)
+    }
+
     fn node_type_uuid(&self) -> [u8; 16] {
         self.as_trait_obj().node_type_uuid()
     }
@@ -1180,9 +1161,10 @@ pub struct CachedCircuitInfo {
     pub is_explicitly_computable: bool,
     pub can_be_sampled: bool,
     pub hash: HashBytes,
-    pub max_non_input_size: BigUint,
+    pub max_non_leaf_size: BigUint,
     pub device_dtype: TorchDeviceDtypeOp,
     pub named_axes: NamedAxes,
+    pub symbolic_size_constraints: SymbolicSizeConstraints,
 }
 
 /// don't want to print hash with Debug, for now just print shape
@@ -1271,22 +1253,33 @@ pub enum ConstructError {
         cumulant_number: usize,
     },
 
-    #[error("len shape different from len axes (len axes: {len_axes}, len shape: {circuit_len_shape}, circuit name: {circuit_name:?})")]
+    #[error("actual_num_children={actual_num_children} != expected_num_children_based_on_axes={expected_num_children_based_on_axes}\neinsum_name={einsum_name:?}")]
+    EinsumWrongNumChildren {
+        actual_num_children: usize,
+        expected_num_children_based_on_axes: usize,
+        einsum_name: Option<String>,
+    },
+
+    #[error("len_axes={len_axes} != child_len_shape={child_len_shape}\neinsum_name={einsum_name:?} child_circuit={child_circuit:?}")]
     EinsumLenShapeDifferentFromAxes {
-        circuit_name: Option<String>,
-        circuit_len_shape: usize,
         len_axes: usize,
+        child_len_shape: usize,
+        einsum_name: Option<String>,
+        child_circuit: CircuitRc,
     },
-    #[error("shape different for axis (axis: {axis}, shape: {circuit_shape}, existing_shape: {existing_shape} circuit name: {circuit_name:?})")]
-    EinsumShapeDifferent {
-        circuit_name: Option<String>,
-        circuit_shape: usize,
+    #[error("new_size={new_size}!=existing_size={existing_size} for axis={axis}\neinsum_name={einsum_name:?}, size introduced with child_circuit={child_circuit:?}")]
+    EinsumAxisSizeDifferent {
+        new_size: SymbolicSizeProduct,
+        existing_size: SymbolicSizeProduct,
         axis: usize,
-        existing_shape: usize,
+        einsum_name: Option<String>,
+        child_circuit: CircuitRc,
     },
-    #[error("output not subset, TODO error")]
+    #[error("circuit_name={circuit_name:?} all_input_axes={all_input_axes:?} output_axes={output_axes:?}")]
     EinsumOutputNotSubset {
-        // TODO: args
+        circuit_name: Option<String>,
+        all_input_axes: HashSet<u8>,
+        output_axes: EinsumAxes,
     },
 
     #[error("Rearrange takes different input shape, shape: {shape:?} spec: {spec:?}")]
@@ -1323,21 +1316,49 @@ pub enum ConstructError {
     #[error("Unknown GeneralFunction name {spec_name}")]
     UnknownGeneralFunction { spec_name: String },
 
-    #[error("ModuleNode wrong number of children, expected {expected} got {got}")]
-    ModuleNodeWrongNumberChildren { expected: usize, got: usize },
+    #[error("Module wrong number of children, expected {expected} got {got}\narg_specs={arg_specs:?} nodes={nodes:?}")]
+    ModuleWrongNumberChildren {
+        expected: usize,
+        got: usize,
+        arg_specs: Vec<ModuleArgSpec>,
+        nodes: Vec<CircuitRc>,
+    },
 
-    #[error("ModuleNode got unknown keyword argument, {argument}")]
-    ModuleNodeUnknownArgument { argument: String },
+    #[error(
+        "Module got unknown keyword argument, {argument}, all_module_inputs={all_module_inputs:?}"
+    )]
+    ModuleUnknownArgument {
+        argument: String,
+        all_module_inputs: Vec<String>,
+    },
 
-    #[error("ModuleNode extract: not all leaves present in circuit, {subcirc:?}")]
-    ModuleNodeExtractNotPresent { subcirc: CircuitRc },
+    #[error("missing_arguments={missing_arguments:?}")]
+    ModuleMissingNames { missing_arguments: Vec<String> },
+
+    #[error("Module extract: not all leaves present in circuit, {subcirc:?}")]
+    ModuleExtractNotPresent { subcirc: CircuitRc },
+
+    #[error("symbols_named_none={symbols_named_none:?}")]
+    ModuleSomeArgsNamedNone { symbols_named_none: Vec<Symbol> },
+
+    #[error("dup_names={dup_names:?}")]
+    ModuleArgsDupNames { dup_names: HashMap<String, usize> },
 
     #[error("spec_circuit={spec_circuit:?} missing_symbols={missing_symbols:?}")]
-    ModuleNodeSomeArgsNotPresent {
+    ModuleSomeArgsNotPresent {
         spec_circuit: CircuitRc,
         missing_symbols: HashSet<Symbol>,
     },
 
+    #[error("actual_circuit={actual_circuit:?}\n(module children are: spec,sym_0,inp_0,sym_1,inp_1,...,sym_n,inp_n)")]
+    ModuleExpectedSymbol { actual_circuit: CircuitRc },
+
+    #[error("orig_module={orig_module:?} got_instead_of_symbol={got_instead_of_symbol:?}")]
+    ModuleExpectedSymbolOnMap {
+        orig_module: Module,
+        got_instead_of_symbol: CircuitRc,
+    },
+
     #[error("Named axis higher than rank")]
     NamedAxisAboveRank {},
 
@@ -1346,17 +1367,34 @@ pub enum ConstructError {
 
     #[error("ndim={ndim}")]
     UnflattenButNDimNot1 { ndim: usize },
+
+    #[error("name_prefix='{name_prefix}' module={module:?}")]
+    ModulePassedNamePrefixAndUseSelfNameAsPrefix { name_prefix: String, module: Module },
 }
 
 pub type CircResult = Result<CircuitRc>;
 
+const INTERNAL_EVAL_ERROR_MESSAGE: &str = "this is likely an internal error (earlier errors should have triggered or an invalid circuit was constructed)";
+
 #[apply(python_error_exception)]
 #[base_error_name(TensorEval)]
 #[base_exception(PyValueError)]
 #[derive(Error, Debug, Clone)]
 pub enum TensorEvalError {
+    #[error(
+        "reached not explicitly computable. {}\ncircuit={circuit:?})",
+        INTERNAL_EVAL_ERROR_MESSAGE
+    )]
+    NotExplicitlyComputableInternal { circuit: CircuitRc },
     #[error("not explicitly computable: {circuit:?})")]
     NotExplicitlyComputable { circuit: CircuitRc },
+    #[error("not constant: {circuit:?})")]
+    NotConstant { circuit: CircuitRc },
+    #[error(
+        "You can't directly evaluate  module. {}\nmodule={module:?})",
+        INTERNAL_EVAL_ERROR_MESSAGE
+    )]
+    ModulesCantBeDirectlyEvalutedInternal { module: Module },
     #[error("incompatible dtype circ:{circ:?} passed:{passed:?}")]
     DeviceDtypeError {
         circ: TorchDeviceDtypeOp,
@@ -1364,31 +1402,53 @@ pub enum TensorEvalError {
     },
 }
 
-#[derive(Clone, Debug, PartialEq, Eq, Hash, PartialOrd, Ord)]
+#[derive(Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
 pub struct CircuitRc(Arc<Circuit>);
 
+impl fmt::Debug for CircuitRc {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        fmt::Debug::fmt(&***self, f)
+    }
+}
+
 pub fn make_children_zero<T: CircuitNode>(circuit: &T) -> T {
     circuit.map_children_unwrap(&mut |child: CircuitRc| {
-        ScalarConstant::new(0.0, child.info().shape.clone(), child.name_cloned()).rc()
+        Scalar::new(0.0, child.info().shape.clone(), child.name_cloned()).rc()
     })
 }
 
-pub fn evaluate_fn(circ: CircuitRc) -> Result<Tensor> {
-    evaluate_fn_dtype_device(circ, Default::default())
+pub fn evaluate(circ: CircuitRc) -> Result<Tensor> {
+    evaluate_device_dtype(circ, Default::default())
 }
 
-pub fn evaluate_fn_dtype_device(
-    circ: CircuitRc,
-    dtype_device: TorchDeviceDtypeOp,
-) -> Result<Tensor> {
-    let device_dtype = dtype_device
+#[pyfunction]
+pub fn check_evaluable(circuit: CircuitRc) -> Result<()> {
+    if !circuit.info().is_explicitly_computable {
+        bail!(TensorEvalError::NotExplicitlyComputable {
+            circuit: circuit.clone()
+        })
+    }
+    if !circuit.info().is_constant {
+        bail!(TensorEvalError::NotConstant {
+            circuit: circuit.clone()
+        })
+    }
+    Ok(())
+}
+
+#[pyfunction(dtype_device = "TorchDeviceDtypeOp::NONE")]
+#[pyo3(name = "evaluate")]
+pub fn evaluate_device_dtype(circ: CircuitRc, device_dtype: TorchDeviceDtypeOp) -> Result<Tensor> {
+    check_evaluable(circ.clone())?;
+    let device_dtype = device_dtype
         .clone()
         .combine(circ.info().device_dtype.clone())
         .map_err(|_err| TensorEvalError::DeviceDtypeError {
             circ: circ.info().device_dtype.clone(),
-            passed: dtype_device.clone(),
+            passed: device_dtype.clone(),
         })?
         .unwrap_or_defaults();
+    let circ = inline_all_modules(circ);
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
     #[use_try]
@@ -1469,31 +1529,48 @@ where
 /// does not visit children of circuits where f fails. It does visit all children even if one fails
 /// even though this is more work than stopping on the first child that fails
 /// because it's semantically cleaner to not have to think about which children are first
-#[apply(pycallable)]
-#[pyo3(name = "visit_circuit")]
-pub fn visit_circuit<F>(circuit: CircuitRc, mut f: F) -> Result<()>
+pub fn visit_circuit_non_free<F>(circuit: CircuitRc, f: F, recur_into_free: bool) -> Result<()>
 where
-    F: FnMut((circuit, CircuitRc)) -> Result<()>,
+    F: FnMut(CircuitRc) -> Result<()>,
 {
     let mut f = f;
     let mut seen: HashSet<HashBytes> = HashSet::default();
 
-    fn recurse<F>(circ: CircuitRc, seen: &mut HashSet<HashBytes>, f: &mut F) -> Result<()>
+    fn recurse<F>(
+        circ: CircuitRc,
+        seen: &mut HashSet<HashBytes>,
+        f: &mut F,
+        recur_into_free: bool,
+    ) -> Result<()>
     where
         F: FnMut(CircuitRc) -> Result<()>,
     {
         if seen.insert(circ.info().hash) {
             f(circ.clone())?;
 
-            circ.children()
-                .map(|child| recurse(child, seen, f))
+            let children = if recur_into_free {
+                circ.children()
+            } else {
+                circ.non_free_children()
+            };
+            children
+                .map(|child| recurse(child, seen, f, recur_into_free))
                 .collect::<Vec<_>>() // intermediate collect causes all recurses to happen even if one errors
                 .into_iter()
                 .collect::<Result<Vec<_>, _>>()?;
         }
         Ok(())
     }
-    recurse(circuit, &mut seen, &mut f)
+    recurse(circuit, &mut seen, &mut f, recur_into_free)
+}
+
+#[apply(pycallable)]
+#[pyo3(name = "visit_circuit")]
+pub fn visit_circuit<F>(circuit: CircuitRc, mut f: F) -> Result<()>
+where
+    F: FnMut((circuit, CircuitRc)) -> Result<()>,
+{
+    visit_circuit_non_free(circuit, f, true)
 }
 
 // TODO: maybe this already exists?
@@ -1748,6 +1825,7 @@ impl CircuitNodeDefer for CircuitRc {
         // deref to avoid infinite recursion
         (**self).as_trait_obj()
     }
+
     fn map_children_enumerate_impl<F>(&self, f: F) -> Result<Self>
     where
         Self: Sized,
@@ -1755,6 +1833,15 @@ impl CircuitNodeDefer for CircuitRc {
     {
         (**self).map_children_enumerate(f).map(CircuitNode::rc)
     }
+    fn map_non_free_children_enumerate_impl<F>(&self, f: F) -> Result<Self>
+    where
+        Self: Sized,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc>,
+    {
+        (**self)
+            .map_non_free_children_enumerate(f)
+            .map(CircuitNode::rc)
+    }
 
     fn custom_c(self) -> Circuit {
         arc_unwrap_or_clone(self.0)
@@ -1822,11 +1909,14 @@ impl PyCircuitBase {
         self.0.intermediate_cost_bound()
     }
 
-    // TODO: probably could be more efficient...
     fn children(&self) -> Vec<CircuitRc> {
         self.0.children().collect()
     }
 
+    fn non_free_children(&self) -> Vec<CircuitRc> {
+        self.0.non_free_children().collect()
+    }
+
     fn __richcmp__(&self, object: &Self, comp_op: CompareOp) -> bool {
         use_rust_comp(&self.0, &object.0, comp_op)
     }
@@ -1858,7 +1948,7 @@ impl PyCircuitBase {
     }
 
     pub fn max_non_input_size(&self) -> BigUint {
-        (*self.0).info().max_non_input_size.clone()
+        (*self.0).info().max_non_leaf_size.clone()
     }
 
     pub fn print_stats(&self) {
@@ -1867,7 +1957,7 @@ impl PyCircuitBase {
 
     #[args(options = "Default::default()")]
     fn repr(&self, options: PrintOptions) -> Result<String> {
-        options.repr_deep((*self.0).crc())
+        options.repr((*self.0).crc())
     }
 
     #[args(options = "Default::default()")]
@@ -1886,9 +1976,8 @@ impl PyCircuitBase {
     fn ndim(&self) -> usize {
         self.rank()
     }
-    #[pyo3(name = "child_axis_map")]
-    fn child_axis_map_py(&self) -> Vec<Vec<Option<usize>>> {
-        self.child_axis_map()
+    fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
+        self.0.child_axis_map()
     }
 
     fn to_py(&self) -> PyObject {
@@ -1897,7 +1986,7 @@ impl PyCircuitBase {
 
     #[args(device_dtype = "Default::default()")]
     fn evaluate(&self, device_dtype: TorchDeviceDtypeOp) -> Result<Tensor> {
-        evaluate_fn_dtype_device(self.0.clone(), device_dtype)
+        evaluate_device_dtype(self.0.clone(), device_dtype)
     }
 
     fn map_children_enumerate(&self, f: PyObject) -> Result<CircuitRc> {
