@@ -8,7 +8,7 @@ use anyhow::{bail, Context, Error, Result};
 use bit_iter::BitIter;
 use circuit_base::{
     computational_node::EinsumArgs, prelude::*, Add, CircuitType, Concat, DiscreteVar, Einsum,
-    GeneralFunction, Index, Rearrange, ScalarConstant, Scatter, Symbol, Tag,
+    GeneralFunction, Index, Rearrange, Scalar, Scatter, Symbol, Tag,
 };
 use itertools::Itertools;
 use macro_rules_attribute::apply;
@@ -61,15 +61,15 @@ pub fn add_flatten_once(add: &Add) -> Option<Add> {
 #[pyfunction]
 pub fn add_collapse_scalar_inputs(add: &Add) -> Option<Add> {
     let nodes = add.nodes.iter().map(|x| (**x).clone());
-    let (scalar_inputs, non_scalar_inputs): (Vec<ScalarConstant>, Vec<Arc<Circuit>>) =
-        filter_by_variant!(nodes, Circuit, ScalarConstant, ScalarConstant);
+    let (scalar_inputs, non_scalar_inputs): (Vec<Scalar>, Vec<Arc<Circuit>>) =
+        filter_by_variant!(nodes, Circuit, Scalar, Scalar);
     if scalar_inputs.len() <= 1 {
         None
     } else {
         let new_scalar: f64 = scalar_inputs.iter().map(|node| node.value).sum();
         let mut new_inputs: Vec<CircuitRc> =
             non_scalar_inputs.iter().map(|x| x.clone().into()).collect();
-        let new_scalar_input = ScalarConstant::new(new_scalar, add.info().shape.clone(), None).rc();
+        let new_scalar_input = Scalar::new(new_scalar, add.info().shape.clone(), None).rc();
         new_inputs.push(new_scalar_input);
         Some(Add::new(new_inputs, add.name_cloned()))
     }
@@ -82,7 +82,7 @@ pub fn add_elim_zeros(add: &Add) -> Option<Add> {
         .nodes
         .iter()
         .filter_map(|node| match &***node {
-            Circuit::ScalarConstant(scalar) => {
+            Circuit::Scalar(scalar) => {
                 if scalar.value == 0.0 && scalar.info().rank() == 0 {
                     did_anything = true;
                     None
@@ -131,7 +131,7 @@ pub fn add_deduplicate(add: &Add) -> Option<Add> {
 #[pyfunction]
 pub fn remove_add_few_input(add: &Add) -> Option<CircuitRc> {
     if add.nodes.is_empty() {
-        Some(ScalarConstant::new(0f64, add.info().shape.clone(), add.name_cloned()).rc())
+        Some(Scalar::new(0f64, add.info().shape.clone(), add.name_cloned()).rc())
     } else if add.nodes.len() == 1 {
         Some(add.nodes[0].clone())
     } else {
@@ -333,7 +333,7 @@ pub fn einsum_nest_optimize(einsum: &Einsum, context: &mut OptimizationContext)
         Some(context.cache.max_single_tensor_numel),
         None,
         Some(filter_to_idx(&einsum.args, &|(x, _)| {
-            matches!(&***x, Circuit::ArrayConstant(_))
+            matches!(&***x, Circuit::Array(_))
         })),
     )?;
     // let path = einsum.get_spec().optimize_dp(None,None,None);
@@ -465,7 +465,7 @@ pub fn einsum_elim_identity(einsum: &Einsum) -> Option<CircuitRc> {
     {
         Some(einsum.args[0].0.clone())
     } else if einsum.args.is_empty() {
-        Some(ScalarConstant::new(1.0, sv![], einsum.name_cloned()).rc())
+        Some(Scalar::new(1.0, sv![], einsum.name_cloned()).rc())
     } else {
         None
     }
@@ -473,8 +473,8 @@ pub fn einsum_elim_identity(einsum: &Einsum) -> Option<CircuitRc> {
 
 #[pyfunction]
 pub fn index_merge_scalar(index: &Index) -> Option<CircuitRc> {
-    index.node.as_scalar_constant().map(|scalar| {
-        ScalarConstant::new(
+    index.node.as_scalar().map(|scalar| {
+        Scalar::new(
             scalar.value,
             index.info().shape.clone(),
             index.name_cloned(),
@@ -505,8 +505,8 @@ pub fn index_fuse(index: &Index) -> Option<Index> {
 
 #[pyfunction]
 pub fn rearrange_merge_scalar(rearrange: &Rearrange) -> Option<CircuitRc> {
-    rearrange.node.as_scalar_constant().map(|scalar| {
-        ScalarConstant::new(
+    rearrange.node.as_scalar().map(|scalar| {
+        Scalar::new(
             scalar.value,
             rearrange.info().shape.clone(),
             rearrange.name_cloned(),
@@ -530,13 +530,13 @@ pub fn concat_merge_uniform(concat: &Concat) -> Option<CircuitRc> {
         .nodes
         .iter()
         .map(|x| match &***x {
-            Circuit::ScalarConstant(scalar) => Some(scalar.value),
+            Circuit::Scalar(scalar) => Some(scalar.value),
             _ => None,
         })
         .collect();
     if maybe_scalars[0].is_some() && maybe_scalars.iter().all(|x| *x == maybe_scalars[0]) {
         Some(
-            ScalarConstant::new(
+            Scalar::new(
                 maybe_scalars[0].unwrap(),
                 concat.info().shape.clone(),
                 concat.name_cloned(),
@@ -559,7 +559,7 @@ pub fn concat_elim_identity(concat: &Concat) -> Option<CircuitRc> {
 
 pub fn get_removable_axes(circuit: &CircuitRc) -> HashSet<usize> {
     match &***circuit {
-        Circuit::ScalarConstant(_scalar) => (0..circuit.info().rank()).collect(),
+        Circuit::Scalar(_scalar) => (0..circuit.info().rank()).collect(),
         Circuit::Rearrange(rearrange) => rearrange
             .spec
             .out_broadcast_axes()
@@ -578,8 +578,8 @@ pub fn remove_axes(circuit: &CircuitRc, axes: &HashSet<usize>) -> Option<Circuit
         t.name_cloned().map(|x| format!("{}_rem_ax", x))
     }
     match &***circuit {
-        Circuit::ScalarConstant(scalar) => Some(
-            ScalarConstant::new(
+        Circuit::Scalar(scalar) => Some(
+            Scalar::new(
                 scalar.value,
                 filter_out_idx(&scalar.info().shape, axes)
                     .into_iter()
@@ -725,7 +725,7 @@ pub fn einsum_pull_removable_axes(einsum: &Einsum) -> Option<CircuitRc> {
         .map(|i| shape_map[i] as f64)
         .product();
     if scalar_mul != 1.0 {
-        new_args.push((ScalarConstant::new(scalar_mul, sv![], None).rc(), sv![]));
+        new_args.push((Scalar::new(scalar_mul, sv![], None).rc(), sv![]));
     }
 
     let new_out_ints: EinsumAxes = einsum
@@ -1021,12 +1021,12 @@ pub fn permute_of_einsum_merge(rearrange: &Rearrange) -> Option<Einsum> {
 }
 
 #[pyfunction]
-pub fn einsum_elim_zero(einsum: &Einsum) -> Option<ScalarConstant> {
+pub fn einsum_elim_zero(einsum: &Einsum) -> Option<Scalar> {
     if einsum.children().any(|x| match &**x {
-        Circuit::ScalarConstant(scalar) => scalar.value == 0.0,
+        Circuit::Scalar(scalar) => scalar.value == 0.0,
         _ => false,
     }) {
-        Some(ScalarConstant::new(
+        Some(Scalar::new(
             0.0,
             einsum.info().shape.clone(),
             einsum.name_cloned(),
@@ -1044,7 +1044,7 @@ pub fn einsum_merge_scalars(einsum: &Einsum) -> Option<Einsum> {
         .args
         .iter()
         .filter(|(node, _ints)| match &***node {
-            Circuit::ScalarConstant(scalar) => {
+            Circuit::Scalar(scalar) => {
                 if scalar.info().shape.is_empty() {
                     num_scalars_found += 1;
                     scalar_mul *= scalar.value;
@@ -1061,7 +1061,7 @@ pub fn einsum_merge_scalars(einsum: &Einsum) -> Option<Einsum> {
         return None;
     }
     if scalar_mul != 1.0 {
-        new_args.push((ScalarConstant::new(scalar_mul, sv![], None).rc(), sv![]))
+        new_args.push((Scalar::new(scalar_mul, sv![], None).rc(), sv![]))
     }
     Some(Einsum::new(new_args, einsum.out_axes.clone(), None))
 }
@@ -1265,6 +1265,9 @@ pub fn push_down_index_raw(
             })
             .flatten()
     };
+    // TODO: maybe handle symbolic axes? Possibilities:
+    // Check we don't index into sum of symbolic axes
+    // Handle rearrange case.
     let out: CircuitRc = match &**node.node {
         Circuit::Add(inner) => {
             let new_operands = inner
@@ -1549,8 +1552,7 @@ pub fn push_down_index_raw(
                     )
                     .unwrap()
                     .extract(py)
-                    .unwrap() // TODO: this used to not be an unwrap - when can this fail!
-            });
+            })?;
             let (index, spec): (TensorIndex, RearrangeSpec) = if let Some(x) = py_out {
                 x
             } else {
@@ -1630,7 +1632,7 @@ pub fn push_down_index_raw(
             var.probs_and_group.clone(),
             out_name,
         ),
-        Circuit::ScalarConstant(_) => index_merge_scalar(node).unwrap(),
+        Circuit::Scalar(_) => index_merge_scalar(node).unwrap(),
         // Circuit::Cumulant(cum) => Cumulant::nrc(
         //     cum.dims_for_self().map(...)
         //     out_name,
@@ -1840,18 +1842,18 @@ pub fn add_fuse_scalar_multiples(add: &Add) -> Option<Add> {
 
     let mut reduced_any = 0;
 
-    let scalars_and_children: Vec<(Einsum, f64, Option<String>)> =
-        add.children()
-            .map(|c| match &**c {
-                Circuit::Einsum(e) => {
-                    let mut scalar: f64 = 1.0;
-                    let mut new_name: Option<String> = None;
-                    let no_scalar_children: EinsumArgs = e
+    let scalars_and_children: Vec<(Einsum, f64, Option<String>)> = add
+        .children()
+        .map(|c| match &**c {
+            Circuit::Einsum(e) => {
+                let mut scalar: f64 = 1.0;
+                let mut new_name: Option<String> = None;
+                let no_scalar_children: EinsumArgs = e
                     .args
                     .iter()
                     .filter(|args| {
                         let (child, _axes) = &**args;
-                        if let Some(x) = child.as_scalar_constant() && x.info().shape.len() == 0  {
+                        if let Some(x) = child.as_scalar() && x.info().shape.len() == 0  {
                             scalar *= x.value;
                             reduced_any += 1;
                             if x.name().is_some() {
@@ -1867,25 +1869,26 @@ pub fn add_fuse_scalar_multiples(add: &Add) -> Option<Add> {
                         }else{
                             true
                         }
-                    }).cloned()
+                    })
+                    .cloned()
                     .collect();
 
-                    (
-                        Einsum::new(no_scalar_children, e.out_axes.clone(), e.name_cloned()),
-                        scalar,
-                        new_name,
-                    )
+                (
+                    Einsum::new(no_scalar_children, e.out_axes.clone(), e.name_cloned()),
+                    scalar,
+                    new_name,
+                )
+            }
+            Circuit::Scalar(s) => {
+                if s.info().shape.len() == 0 {
+                    (Einsum::empty(None), s.value, s.name_cloned())
+                } else {
+                    (Einsum::identity(c, None), 1.0, None)
                 }
-                Circuit::ScalarConstant(s) => {
-                    if s.info().shape.len() == 0 {
-                        (Einsum::empty(None), s.value, s.name_cloned())
-                    } else {
-                        (Einsum::identity(c, None), 1.0, None)
-                    }
-                }
-                _ => (Einsum::identity(c, None), 1.0, None),
-            })
-            .collect();
+            }
+            _ => (Einsum::identity(c, None), 1.0, None),
+        })
+        .collect();
 
     if reduced_any <= 1 {
         return None;
@@ -1913,7 +1916,7 @@ pub fn add_fuse_scalar_multiples(add: &Add) -> Option<Add> {
                 let name = einsum.name_cloned();
                 let out_axes = einsum.out_axes;
                 let mut v = einsum.args;
-                v.push((ScalarConstant::new(scalar, sv![], scalar_name).rc(), sv![]));
+                v.push((Scalar::new(scalar, sv![], scalar_name).rc(), sv![]));
                 Einsum::nrc(v, out_axes, name)
             }
         })
