@@ -5,8 +5,8 @@ use circuit_base::{
     deep_map_op, deep_map_unwrap,
     named_axes::{axis_of_name, get_axis_leaves, set_named_axes},
     prelude::*,
-    visit_circuit_unwrap, visit_circuit_with_parents, Concat, Einsum, Index, ModuleNode,
-    ModuleNodeArgSpec, ModuleNodeSpec, Symbol,
+    visit_circuit_unwrap, visit_circuit_with_parents, Concat, Einsum, Index, Module, ModuleArgSpec,
+    ModuleSpec, Symbol,
 };
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
@@ -46,12 +46,12 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
     // for k in &leaves_non_axis {
     //     k.print();
     // }
-    let input_specs_axis: Vec<(CircuitRc, ModuleNodeArgSpec)> = leaves_axis
+    let input_specs_axis: Vec<(CircuitRc, ModuleArgSpec)> = leaves_axis
         .iter()
         .map(|(sub, i)| {
             (
                 sub.clone(),
-                ModuleNodeArgSpec {
+                ModuleArgSpec {
                     symbol: set_named_axes(
                         Symbol::new_with_random_uuid(sub.info().shape.clone(), None),
                         BTreeMap::from([(*i as u8, name.clone())]),
@@ -62,13 +62,13 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
             )
         })
         .collect();
-    let input_specs: Vec<(CircuitRc, ModuleNodeArgSpec)> = input_specs_axis
+    let input_specs: Vec<(CircuitRc, ModuleArgSpec)> = input_specs_axis
         .iter()
         .cloned()
         .chain(leaves_non_axis.iter().map(|sub| {
             (
                 sub.clone(),
-                ModuleNodeArgSpec {
+                ModuleArgSpec {
                     symbol: Symbol::new_with_random_uuid(sub.info().shape.clone(), None),
                     batchable: true,
                     expandable: true,
@@ -83,13 +83,19 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
     if input_specs.is_empty() {
         bail!(BatchError::AxisOriginatesTooHigh {});
     }
-    let module_spec =
-        ModuleNodeSpec::new_extract(circuit.clone(), input_specs.clone(), None, false)?;
+    let require_all_args = false; // this really should be true - get_axis_leaves doesn't get leaves!
+    let module_spec = ModuleSpec::new_extract(
+        circuit.clone(),
+        input_specs.clone(),
+        require_all_args,
+        false,
+    )
+    .unwrap();
     // println!("module spec");
     // module_spec.spec_circuit.print();
     let module_spec = module_spec.resize(
         module_spec
-            .input_specs
+            .arg_specs
             .iter()
             .map(|inp_spec| {
                 if let Some((c, _is)) = input_specs_axis
@@ -109,7 +115,7 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
     let concattands: Vec<CircuitRc> = (0..num_batches)
         .map(|i| {
             let new_nodes: Vec<CircuitRc> = module_spec
-                .input_specs
+                .arg_specs
                 .iter()
                 .map(|inp_spec| {
                     if let Some((c, _is)) = input_specs_axis
@@ -142,7 +148,7 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
             // for c in new_nodes.iter() {
             //     c.print();
             // }
-            ModuleNode::nrc(new_nodes, module_spec.clone(), None)
+            Module::nrc(new_nodes, module_spec.clone(), None)
         })
         .collect();
     let result = Concat::nrc(concattands, axis, None);
