@@ -2,8 +2,7 @@ use std::io::prelude::*;
 
 use anyhow::{bail, Context, Result};
 use circuit_base::{
-    circuit_utils::has_leaf_symbols, deep_map_op_context, flat_concat, get_compatible_dtype,
-    print::PrintOptions, CircuitNode, CircuitRc, ModuleNode, ModuleNodeSpec,
+    flat_concat, get_compatible_dtype, print::PrintOptions, CircuitNode, CircuitRc, ModuleSpec,
 };
 use num_bigint::BigUint;
 use pyo3::prelude::*;
@@ -174,10 +173,9 @@ pub struct OptimizationCache {
     pub flops: HashMap<HashBytes, BigUint>,
     pub sum_of_node_sizes: HashMap<HashBytes, BigUint>,
     pub canonicalized: HashMap<HashBytes, Option<CircuitRc>>,
-    pub module_specs_optimized_same_settings: HashMap<ModuleNodeSpec, ModuleNodeSpec>,
-    pub module_specs_scheduled_same_settings: HashMap<ModuleNodeSpec, Schedule>,
-    pub module_specs_expanded_shape: HashMap<(ModuleNodeSpec, Vec<Shape>), CircuitRc>,
-    pub module_specs_expanded: HashMap<(ModuleNodeSpec, Vec<HashBytes>), CircuitRc>,
+    pub module_specs_scheduled_same_settings: HashMap<ModuleSpec, Schedule>,
+    pub module_specs_expanded_shape: HashMap<(ModuleSpec, Vec<Shape>), CircuitRc>,
+    pub module_specs_expanded: HashMap<(ModuleSpec, Vec<HashBytes>), CircuitRc>,
     pub stripped_names: HashMap<HashBytes, Option<CircuitRc>>,
     pub times_distributed: usize,
     pub simplification_log: Vec<&'static str>,
@@ -241,21 +239,13 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
     let print_timings = context.settings.verbose >= 2;
     let circuit = timed!(strip_names(circuit, context), 10, print_timings);
     let circuit = remove_autotags(circuit);
-    let circuit = timed!(
-        deep_optimize_module_specs(circuit, context),
-        10,
-        print_timings
-    );
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 4 {
         python_println!("Original Circuit");
         circuit.printu();
     }
     if context.settings.save_circ {
-        let s = PrintOptions::new(true, true, None, false, false)
-            .unwrap()
-            .repr_deep(circuit.clone())
-            .expect("can't fail with None traversal");
+        let s = circuit.repru();
         let f_name = format!("cached_circ_{}.circ", circuit.info().hash_usize());
         python_println!("Save circ to {}", f_name);
 
@@ -277,7 +267,9 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         python_println!("Simplified Circuit");
-        circuit.printu();
+        PrintOptions::compiler_default()
+            .print(circuit.clone())
+            .unwrap();
     }
     let circuit = timed!(
         apply_fn_until_same(&circuit, |x| {
@@ -291,7 +283,9 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         python_println!("Distributed Circuit");
-        circuit.printu();
+        PrintOptions::compiler_default()
+            .print(circuit.clone())
+            .unwrap();
     }
     let circuit = timed!(deep_heuristic_nest_adds(circuit), 10, print_timings);
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
@@ -301,7 +295,9 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         python_println!("Final Circuit");
-        circuit.printu();
+        PrintOptions::compiler_default()
+            .print(circuit.clone())
+            .unwrap();
     }
     circuit
 }
@@ -317,8 +313,7 @@ pub fn optimize_to_schedule(
         python_println!("Optimizing verbose {}", verbose)
     }
     if settings.capture_and_print {
-        PrintOptions::new(true, false, None, false, false)
-            .unwrap()
+        PrintOptions::compiler_default()
             .print(circuit.clone())
             .context("capture print fail!")?;
     }
@@ -341,8 +336,8 @@ pub fn optimize_to_schedule(
 
 #[pyfunction]
 pub fn optimize_and_evaluate(circuit: CircuitRc, settings: OptimizationSettings) -> Result<Tensor> {
-    if has_leaf_symbols(circuit.clone(), &Default::default()) {
-        bail!(SchedulingError::LeafSymbols {});
+    if !circuit.info().is_explicitly_computable {
+        bail!(SchedulingError::NotExplicitlyComputable { circuit });
     }
     let schedule = optimize_to_schedule(circuit, settings)?;
 
@@ -379,47 +374,3 @@ pub fn optimize_to_schedule_many(
     );
     Ok(schedule)
 }
-
-pub fn optimize_module_spec(
-    spec: ModuleNodeSpec,
-    context: &mut OptimizationContext,
-) -> ModuleNodeSpec {
-    context
-        .cache
-        .module_specs_optimized_same_settings
-        .get(&spec)
-        .cloned()
-        .unwrap_or_else(|| {
-            let result = ModuleNodeSpec {
-                input_specs: spec.input_specs.clone(),
-                name: spec.name.clone(),
-                spec_circuit: optimize_circuit(spec.spec_circuit.clone(), context),
-            };
-            context
-                .cache
-                .module_specs_optimized_same_settings
-                .insert(spec, result.clone());
-            result
-        })
-}
-
-pub fn deep_optimize_module_specs(
-    circuit: CircuitRc,
-    context: &mut OptimizationContext,
-) -> CircuitRc {
-    deep_map_op_context(
-        circuit.clone(),
-        &|circuit, context| {
-            circuit.as_module_node().map(|mn| {
-                ModuleNode::nrc(
-                    mn.nodes.clone(),
-                    optimize_module_spec(mn.spec.clone(), context),
-                    mn.name_cloned(),
-                )
-            })
-        },
-        context,
-        &mut Default::default(),
-    )
-    .unwrap_or(circuit)
-}
