@@ -9,9 +9,9 @@ use num_bigint::BigUint;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
     python_error_exception, timed, timed_value,
-    util::{apply_fn_until_none, apply_fn_until_same, mapping_until_end, AsOp},
+    util::{apply_fn_until_none, apply_fn_until_same, mapping_until_end, AsOp, HashBytes},
 };
-use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
+use rustc_hash::FxHashMap as HashMap;
 use thiserror::Error;
 
 use crate::{
@@ -201,7 +201,7 @@ fn all_simp_fns() -> SimpFns {
             f_wrap!(scatter_fuse),
             f_wrap!(scatter_pull_removable_axes),
         ],
-        module_node: vec![l_wrap!(elim_empty_module)],
+        module: vec![l_wrap!(elim_empty_module)],
         ..Default::default()
     }
 }
@@ -548,62 +548,116 @@ pub fn deep_push_down_index_raw(circ: CircuitRc, min_size: Option<usize>) -> Cir
     })
 }
 
+#[derive(Debug, Clone, Default)]
+pub struct CircBloomFilter {
+    pub bits: u128,
+}
+impl CircBloomFilter {
+    #[inline]
+    pub fn insert(&mut self, x: usize) {
+        self.bits |= 1 << (x & 0b1111111)
+    }
+    #[inline]
+    pub fn contains(&mut self, x: usize) -> bool {
+        self.bits & 1 << (x & 0b1111111) != 0
+    }
+    #[inline]
+    pub fn is_superset(&self, other: &CircBloomFilter) -> bool {
+        other.bits & !self.bits == 0
+    }
+    #[inline]
+    pub fn intersection_count(&self, other: &CircBloomFilter) -> u32 {
+        (other.bits & self.bits).count_ones()
+    }
+}
+
 /// we want adds to be nested rather than flat so arguments can be dropped if they're only needed
 /// in future adds
 /// this is suboptimal in many ways. one is broadcasts allow outer products which should be avoided but aren't
 /// for each add, greedily nest into preexisting adds
+fn child_bloom_filter(add: &Add) -> CircBloomFilter {
+    let mut result: CircBloomFilter = Default::default();
+    for node in &add.nodes {
+        result.insert(node.info().hash_usize());
+    }
+    result
+}
 #[pyfunction]
 pub fn deep_heuristic_nest_adds(circ: CircuitRc) -> CircuitRc {
     let circ = deep_canonicalize(circ, &mut Default::default());
-    let mut seen_adds: HashSet<Add> = HashSet::default();
+    let mut seen_adds: Vec<(Add, CircBloomFilter)> = vec![];
     visit_circuit_unwrap(circ.clone(), |c: CircuitRc| {
         if let Some(add) = c.as_add() {
-            seen_adds.insert(add.clone());
+            seen_adds.push((add.clone(), child_bloom_filter(add)));
         }
     });
     // TODO: profile and optimize these
 
-    let mut intersections: HashSet<Add> = HashSet::default();
+    let mut intersections: HashMap<Add, CircBloomFilter> = HashMap::default();
     for circ in &seen_adds {
         for circ2 in &seen_adds {
-            let intersection = Add::new(
-                circ.nodes
-                    .iter()
-                    .filter(|x| circ2.nodes.contains(x))
-                    .cloned()
-                    .collect(),
-                None,
-            );
-            if intersection.nodes.len() >= 2 && &intersection != circ && &intersection != circ2 {
-                intersections.insert(intersection);
+            if circ.1.intersection_count(&circ2.1) >= 2 && circ.0 != circ2.0 {
+                let intersection = Add::new(
+                    circ.0
+                        .nodes
+                        .iter()
+                        .filter(|x| circ2.0.nodes.contains(x))
+                        .cloned()
+                        .collect(),
+                    None,
+                );
+                if intersection.nodes.len() >= 2
+                    && &intersection != &circ.0
+                    && &intersection != &circ2.0
+                {
+                    intersections.insert(intersection.clone(), child_bloom_filter(&intersection));
+                }
             }
         }
     }
     seen_adds.extend(intersections);
+    let mut seen_adds_index: HashMap<HashBytes, usize> = seen_adds
+        .iter()
+        .enumerate()
+        .map(|(i, (c, _b))| (c.info().hash, i))
+        .collect();
     let mut mapping: HashMap<Add, Add> = HashMap::default();
-    while let Some((sup, new_sup)) = (|| {
-        for cand_sub in &seen_adds {
-            if cand_sub.nodes.len() >= 2 {
-                for cand_sup in &seen_adds {
-                    if cand_sub != cand_sup
-                        && cand_sup
-                            .nodes
-                            .iter()
-                            .position(|x| x.info().hash == cand_sub.info().hash)
-                            .is_some()
+    for _ in 0..3 {
+        let mut to_do: Vec<((Add, CircBloomFilter), (Add, CircBloomFilter))> = vec![];
+        for cand_sup in &seen_adds {
+            if cand_sup.0.nodes.len() >= 3 {
+                for cand_sub in &seen_adds {
+                    if cand_sup.1.is_superset(&cand_sub.1)
+                        && cand_sub.0.nodes.len() < cand_sup.0.nodes.len()
+                        && cand_sub.0 != cand_sup.0
                     {
-                        if let Some(new) = extract_add(cand_sup, cand_sub) {
-                            return Some((cand_sup.clone(), new));
+                        if let Some(new) = extract_add(&cand_sup.0, &cand_sub.0) {
+                            let new_entry = (new.clone(), child_bloom_filter(&new));
+                            to_do.push(((cand_sup.0.clone(), cand_sup.1.clone()), new_entry));
                         }
                     }
                 }
             }
         }
-        None
-    })() {
-        seen_adds.remove(&sup);
-        seen_adds.insert(new_sup.clone());
-        mapping.insert(sup, new_sup.clone());
+        // println!("nest op counts bloom {} extract {} join {}",bloom_count,extract_count,join_count);
+        if to_do.is_empty() {
+            break;
+        }
+        for (sup, new_sup) in to_do {
+            let mut seen_index = seen_adds_index.remove(&sup.0.info().hash);
+            let to_insert = (new_sup.0.clone(), new_sup.1.clone());
+            if let Some(si) = seen_index {
+                seen_adds[si] = to_insert;
+            } else {
+                seen_index = Some(seen_adds.len());
+                seen_adds.push(to_insert);
+            }
+            if !seen_adds_index.contains_key(&new_sup.0.info().hash) {
+                seen_adds_index.insert(new_sup.0.info().hash, seen_index.unwrap());
+            }
+            mapping.insert(sup.0.clone(), new_sup.0.clone());
+        }
+        // break
     }
 
     deep_map_preorder_unwrap(circ, |c| {
