@@ -2,47 +2,50 @@ use std::iter::zip;
 
 use anyhow::{anyhow, bail, Result};
 use circuit_base::{
-    circuit_utils::count_nodes, CircuitNode, CircuitRc, Concat, ModuleNode, ModuleNodeArgSpec,
-    ModuleNodeSpec,
+    circuit_utils::count_nodes, CircuitNode, CircuitRc, Concat, Module, ModuleArgSpec, ModuleSpec,
 };
 use pyo3::prelude::*;
 use rr_util::py_types::PyOpAtAxes;
 
 use crate::{circuit_manipulation::replace_nodes_py, concat_rewrite::split_to_concat};
+
 #[pyfunction]
-pub fn elim_empty_module(circuit: &ModuleNode) -> Option<CircuitRc> {
-    if count_nodes(circuit.spec.spec_circuit.clone()) == 1 {
-        return Some(circuit.nodes[0].clone());
-    }
-    return None;
+pub fn elim_empty_module(circuit: &Module) -> Option<CircuitRc> {
+    (count_nodes(circuit.spec.circuit.clone()) == 1).then(|| circuit.substitute(None, None))
 }
 
-#[pyfunction(require_all_inputs = "false")]
+#[pyfunction(
+    prefix_to_strip = "None",
+    module_name = "None",
+    require_all_args = "true",
+    check_unique_arg_names = "true"
+)]
 pub fn extract_rewrite_raw(
     circuit: CircuitRc,
-    input_specs: Vec<(CircuitRc, ModuleNodeArgSpec)>,
+    input_specs: Vec<(CircuitRc, ModuleArgSpec)>,
     prefix_to_strip: Option<String>,
     module_name: Option<String>,
-    spec_name: Option<String>,
-    require_all_inputs: bool,
-) -> Result<ModuleNode> {
-    let mut spec =
-        ModuleNodeSpec::new_extract(circuit, input_specs.clone(), spec_name, require_all_inputs)?;
+    require_all_args: bool,
+    check_unique_arg_names: bool,
+) -> Result<Module> {
+    let mut spec = ModuleSpec::new_extract(
+        circuit,
+        input_specs.clone(),
+        require_all_args,
+        check_unique_arg_names,
+    )?;
     if let Some(pref) = &prefix_to_strip {
-        spec = spec.map_circuit_unwrap(
-            |x| {
-                if let Some(name) = x.name() {
-                    x.clone()
-                        .rename(Some(name.strip_prefix(pref).unwrap_or(name).to_owned()))
-                } else {
-                    x
-                }
-            },
-            None,
-        )
+        spec = spec.map_circuit_unwrap(|x| {
+            if let Some(name) = x.name() {
+                x.clone()
+                    .rename(Some(name.strip_prefix(pref).unwrap_or(name).to_owned()))
+            } else {
+                x
+            }
+        })
     }
-    ModuleNode::try_new(
-        spec.input_specs
+    Module::try_new(
+        spec.arg_specs
             .iter()
             .map(|x| input_specs.iter().find(|z| &z.1 == x).unwrap().0.clone())
             .collect(),
@@ -55,7 +58,7 @@ pub fn extract_rewrite_raw(
 #[pyfunction]
 pub fn fuse_concat_modules(
     circuit: CircuitRc,
-    modules: Vec<ModuleNode>,
+    modules: Vec<Module>,
     name: Option<String>,
 ) -> Result<CircuitRc> {
     if modules.is_empty() {
@@ -65,14 +68,14 @@ pub fn fuse_concat_modules(
     if !modules.iter().all(|x| x.spec == spec) {
         bail!(anyhow!("modules_one_batch not all same"))
     }
-    let pure_rank = spec.spec_circuit.info().rank();
+    let pure_rank = spec.circuit.info().rank();
     let modules_one_batch = modules
         .iter()
         .map(|x| {
-            let dif = x.info().rank() - spec.spec_circuit.info().rank();
+            let dif = x.info().rank() - spec.circuit.info().rank();
             match dif {
-                0 => Ok(ModuleNode::new(
-                    zip(&x.nodes, &x.spec.input_specs)
+                0 => Ok(Module::new(
+                    zip(&x.nodes, &x.spec.arg_specs)
                         .map(|(n, s)| {
                             if s.batchable {
                                 n.unsqueeze(vec![0], None).unwrap().rc()
@@ -91,7 +94,7 @@ pub fn fuse_concat_modules(
         .collect::<Result<Vec<_>>>()?;
     if !modules_one_batch[0]
         .spec
-        .input_specs
+        .arg_specs
         .iter()
         .enumerate()
         .all(|(i, x)| {
@@ -105,10 +108,10 @@ pub fn fuse_concat_modules(
             "all inputs must be batchable or have the same child on all inputs"
         ));
     }
-    let concatted_input = ModuleNode::nrc(
-        (0..spec.input_specs.len())
+    let concatted_input = Module::nrc(
+        (0..spec.arg_specs.len())
             .map(|i| {
-                if spec.input_specs[i].batchable {
+                if spec.arg_specs[i].batchable {
                     Concat::nrc(
                         modules_one_batch
                             .iter()
