@@ -2,13 +2,12 @@ use std::fmt::{self, Display};
 
 use anyhow::{bail, Context, Result};
 use circuit_base::{
-    circuit_utils::{hash_to_node, toposort_circuit},
+    circuit_utils::{hash_to_node_non_free, toposort_circuit},
     get_compatible_dtype,
-    module_nodes::inline_all_modules,
+    module::inline_all_modules,
     prelude::*,
-    print::oom_fmt,
-    visit_circuit_unwrap, Add, ArrayConstant, IrreducibleNode, Leaf, ModuleNodeSpec,
-    ScalarConstant, Symbol,
+    print::{oom_fmt, PrintOptions},
+    visit_circuit_non_free, Add, Array, IrreducibleNode, Leaf, ModuleSpec, Scalar, Symbol,
 };
 use itertools::Itertools;
 use macro_rules_attribute::apply;
@@ -63,7 +62,7 @@ pub struct Schedule {
     // before when these were in tensors, they had to be right dtype and therefore overflowed when
     // adjustment needed
     #[pyo3(get)]
-    pub scalar_constants: HashMap<usize, ScalarConstant>,
+    pub scalars: HashMap<usize, Scalar>,
     #[pyo3(get)]
     pub device_dtype: TorchDeviceDtype,
     pub output_circuit: Option<(usize, CircuitRc)>,
@@ -81,10 +80,9 @@ impl Schedule {
                     "key circuit wasn't present in original",
                 ));
             }
-            result.constants.insert(
-                result.old_constant_hashes[&k],
-                ArrayConstant::new(v, None).into(),
-            );
+            result
+                .constants
+                .insert(result.old_constant_hashes[&k], Array::new(v, None).into());
         }
         Ok(result)
     }
@@ -97,9 +95,7 @@ impl Schedule {
             let maybe_tensor: Option<Tensor> =
                 Python::with_gil(|py| pycall!(f, (PyBytes::new(py, key),)));
             if let Some(t) = maybe_tensor {
-                result
-                    .constants
-                    .insert(*id, ArrayConstant::new(t, None).into());
+                result.constants.insert(*id, Array::new(t, None).into());
             }
         });
         Ok(result)
@@ -177,7 +173,7 @@ impl Schedule {
         *self
             .constants
             .keys()
-            .chain(self.scalar_constants.keys())
+            .chain(self.scalars.keys())
             .chain(self.instructions.iter().filter_map(|ins| match ins {
                 Instruction::Compute(k, _c) => Some(k),
                 _ => None,
@@ -258,7 +254,7 @@ impl Display for Schedule {
                 .join("\n"),
             self.constants
                 .iter()
-                .filter_map(|(k, c)| c.as_array_constant().map(|_| k.to_string()))
+                .filter_map(|(k, c)| c.as_array().map(|_| k.to_string()))
                 .collect::<Vec<String>>()
                 .join(" "),
             self.constants
@@ -266,7 +262,7 @@ impl Display for Schedule {
                 .filter_map(|(k, c)| c.as_symbol().map(|_| k.to_string()))
                 .collect::<Vec<String>>()
                 .join(" "),
-            self.scalar_constants
+            self.scalars
                 .iter()
                 .map(|(k, _c)| k.to_string())
                 .collect::<Vec<String>>()
@@ -307,8 +303,8 @@ pub enum SchedulingOOMError {
 #[base_exception(PyValueError)]
 #[derive(Error, Debug)]
 pub enum SchedulingError {
-    #[error("Circuit has leaf symbols")]
-    LeafSymbols {},
+    #[error("circuit={circuit:?}")]
+    NotExplicitlyComputable { circuit: CircuitRc },
 }
 
 impl Display for ScheduleStats {
@@ -371,12 +367,12 @@ pub fn evaluate_schedule(schedule: &Schedule) -> HashMap<usize, Tensor> {
     let mut live: HashMap<usize, Tensor> = schedule
         .constants
         .iter()
-        .map(|x| x.1.as_array_constant().map(|z| (*x.0, z.value.clone())))
+        .map(|x| x.1.as_array().map(|z| (*x.0, z.value.clone())))
         .collect::<Option<HashMap<_, _>>>()
         .unwrap();
     live.extend(
         schedule
-            .scalar_constants
+            .scalars
             .iter()
             .map(|x| (*x.0, x.1.eval_tensors(&[], &schedule.device_dtype).unwrap())),
     );
@@ -455,14 +451,9 @@ pub fn evaluate_schedule_adjust_numerical_scale(
     let mut live: HashMap<usize, (Tensor, f64)> = schedule
         .constants
         .iter()
-        .map(|(key, x)| {
-            (
-                *key,
-                clamp(&(x.as_array_constant().unwrap().value.clone(), 1.0)),
-            )
-        })
+        .map(|(key, x)| (*key, clamp(&(x.as_array().unwrap().value.clone(), 1.0))))
         .collect();
-    live.extend(schedule.scalar_constants.iter().map(|(h, s)| {
+    live.extend(schedule.scalars.iter().map(|(h, s)| {
         (*h, {
             let value_scale = s.value.abs();
             if value_scale > settings.numerical_scale_max
@@ -547,7 +538,7 @@ pub fn evaluate_schedule_adjust_numerical_scale(
                         .unwrap(),
                         tensors_and_scales[0].1,
                     ),
-                    Circuit::ScalarConstant(_) | Circuit::ArrayConstant(_) | Circuit::Symbol(_) => {
+                    Circuit::Scalar(_) | Circuit::Array(_) | Circuit::Symbol(_) => {
                         panic!("constant found as schedule instruction, not supposed to happen")
                     }
                     _ => {
@@ -572,7 +563,7 @@ pub fn evaluate_schedule_adjust_numerical_scale(
 pub fn order_to_schedule(
     order: &Vec<CircuitRc>,
     constants: &Vec<IrreducibleNode>,
-    scalar_constants: &Vec<ScalarConstant>,
+    scalars: &Vec<Scalar>,
     to_keep: HashSet<HashBytes>,
     device_dtype: TorchDeviceDtype,
 ) -> (Schedule, Vec<usize>) {
@@ -587,7 +578,7 @@ pub fn order_to_schedule(
     let mut result: Schedule = Schedule {
         instructions: vec![],
         constants: constants.clone(),
-        scalar_constants: scalar_constants
+        scalars: scalars
             .iter()
             .map(|x| {
                 circ_to_id.insert(x.info().hash, circ_to_id.len());
@@ -618,7 +609,7 @@ pub fn order_to_schedule(
         let next_id = circ_to_id.len();
         let our_id = *circ_to_id.entry(ex.info().hash).or_insert(next_id);
         let node_here_symbol_children = ex
-            .map_children_unwrap(|child| {
+            .map_non_free_children_unwrap(|child| {
                 child_from_key(circ_to_id[&child.info().hash], child.info().shape.clone()).rc()
             })
             .rc();
@@ -648,28 +639,40 @@ pub fn circuits_to_dag(circuits: &[CircuitRc]) -> Dag {
         (result.node_hashes.len() - 1) as u32
     };
     for circuit in circuits {
-        visit_circuit_unwrap(circuit.clone(), |c: CircuitRc| {
-            // ArrayConstants are never added to the dag to begin with because they're always 0 cost
-            if !Leaf::matches(&c) {
-                let my_number: u32 = number_node(c.clone(), &mut result);
-                let children_to_consider: Vec<CircuitRc> =
-                    c.children().filter(|child| !Leaf::matches(child)).collect();
-                result.children[my_number as usize] = children_to_consider
-                    .iter()
-                    .map(|child| number_node(child.clone(), &mut result))
-                    .unique()
-                    .collect();
-                for child in children_to_consider {
-                    let new_num = number_node(child.clone(), &mut result);
-                    if !result.parents[new_num as usize]
+        // non free so we avoid recuring into module spec.circuit
+        visit_circuit_non_free(
+            circuit.clone(),
+            |c: CircuitRc| {
+                // Arrays are never added to the dag to begin with because they're always 0 cost
+
+                if !Leaf::matches(&c) {
+                    let my_number: u32 = number_node(c.clone(), &mut result);
+
+                    let children_to_consider: Vec<CircuitRc> = c
+                        .non_free_children()
+                        .filter(|child| !Leaf::matches(child))
+                        .collect();
+                    result.children[my_number as usize] = children_to_consider
                         .iter()
-                        .contains(&(my_number as u32))
-                    {
-                        result.parents[new_num as usize].push(my_number as u32);
+                        .map(|child| number_node(child.clone(), &mut result))
+                        .unique()
+                        .collect();
+                    for child in children_to_consider {
+                        let new_num = number_node(child.clone(), &mut result);
+                        if !result.parents[new_num as usize]
+                            .iter()
+                            .contains(&(my_number as u32))
+                        {
+                            result.parents[new_num as usize].push(my_number as u32);
+                        }
                     }
                 }
-            }
-        });
+
+                Ok(())
+            },
+            false,
+        )
+        .unwrap();
     }
     result.node_to_orig = (0..result.node_costs.len() as u32)
         .map(|x| (x, sv![x].clone()))
@@ -682,9 +685,9 @@ pub fn circuit_to_schedule(
     context: &mut OptimizationContext,
 ) -> Result<Schedule> {
     let mut dag = circuits_to_dag(&[circuit.clone()]);
-    if circuit.info().max_non_input_size > BigUint::from(context.cache.fallback_total_numel) {
+    if circuit.info().max_non_leaf_size > BigUint::from(context.cache.fallback_total_numel) {
         bail!(SchedulingOOMError::Single {
-            numel: circuit.info().max_non_input_size.clone(),
+            numel: circuit.info().max_non_leaf_size.clone(),
         });
     }
     let order_result = {
@@ -722,11 +725,11 @@ pub fn circuit_to_schedule(
     let order = with_context_failable(order_result, || {
         Ok(format!(
             "getting schedule failed for circuit:\n{}",
-            circuit.repr()?
+            PrintOptions::compiler_default().repr(circuit.clone())?
         ))
     })?;
 
-    let to_node = hash_to_node(circuit.clone());
+    let to_node = hash_to_node_non_free(circuit.clone(), false);
     let circuit_order: Vec<CircuitRc> = order
         .iter()
         .map(|x| to_node[&dag.node_hashes[*x as usize]].clone())
@@ -739,7 +742,7 @@ pub fn circuit_to_schedule(
             .collect(),
         &to_node
             .iter()
-            .filter_map(|x| x.1.as_scalar_constant().cloned())
+            .filter_map(|x| x.1.as_scalar().cloned())
             .collect(),
         HashSet::from_iter([circuit.info().hash]),
         get_compatible_dtype(&circuit),
@@ -765,7 +768,6 @@ pub fn circuit_to_schedule_naive_toposort(circuit: CircuitRc) -> Schedule {
 
 #[pyfunction]
 pub fn scheduled_evaluate(circuit: CircuitRc, settings: OptimizationSettings) -> Result<Tensor> {
-    if settings.scheduling_naive {}
     let schedule = if settings.scheduling_naive {
         Ok(circuit_to_schedule_naive_toposort(circuit))
     } else {
@@ -777,14 +779,14 @@ pub fn scheduled_evaluate(circuit: CircuitRc, settings: OptimizationSettings) ->
     Ok(evaluate_schedule(&schedule)[&schedule.output_circuit.unwrap().0].clone())
 }
 
-fn schedule_module_spec(spec: &ModuleNodeSpec, context: &mut OptimizationContext) -> Schedule {
+fn schedule_module_spec(spec: &ModuleSpec, context: &mut OptimizationContext) -> Schedule {
     context
         .cache
         .module_specs_scheduled_same_settings
         .get(spec)
         .cloned()
         .unwrap_or_else(|| {
-            let result = circuit_to_schedule(spec.spec_circuit.clone(), context).unwrap();
+            let result = circuit_to_schedule(spec.circuit.clone(), context).unwrap();
             context
                 .cache
                 .module_specs_scheduled_same_settings
@@ -795,8 +797,8 @@ fn schedule_module_spec(spec: &ModuleNodeSpec, context: &mut OptimizationContext
 
 fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationContext) -> Schedule {
     let mut result = schedule.clone();
-    let mut scalar_to_id: HashMap<ScalarConstant, usize> = schedule
-        .scalar_constants
+    let mut scalar_to_id: HashMap<Scalar, usize> = schedule
+        .scalars
         .iter()
         .map(|(a, b)| (b.clone(), *a))
         .collect();
@@ -811,8 +813,9 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
         match ins {
             Instruction::Drop(_d) => result.instructions.push(ins.clone()),
             Instruction::Compute(outermost_k, v) => match &***v {
-                Circuit::ModuleNode(mn) => {
-                    let original_children: Vec<usize> = mn.children().map(get_child_key).collect();
+                Circuit::Module(mn) => {
+                    let original_children: Vec<usize> =
+                        mn.nodes.iter().cloned().map(get_child_key).collect();
 
                     let inner_schedule = schedule_module_spec(&mn.spec, context);
                     if inner_schedule.instructions.is_empty() {
@@ -828,10 +831,10 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
                             inner_schedule.output_circuit.clone().unwrap().0,
                             *outermost_k,
                         )]);
-                        for (inner_k, sc) in &inner_schedule.scalar_constants {
+                        for (inner_k, sc) in &inner_schedule.scalars {
                             let outer = scalar_to_id.get(sc).cloned().unwrap_or_else(|| {
                                 scalar_to_id.insert(sc.clone(), next_key);
-                                result.scalar_constants.insert(next_key, sc.clone());
+                                result.scalars.insert(next_key, sc.clone());
                                 next_key += 1;
                                 next_key - 1
                             });
@@ -844,7 +847,7 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
                                     .cloned()
                                     .unwrap_or_else(|| {
                                         if let Some(pos) =
-                                            mn.spec.input_specs.iter().position(|argspec| {
+                                            mn.spec.arg_specs.iter().position(|argspec| {
                                                 argspec.symbol.info().hash
                                                     == irreducible.info().hash
                                             })
@@ -894,6 +897,7 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
                         }
                     }
                 }
+                // Note: if other node types have free we'd have to handle that here!
                 _ => {
                     result.instructions.push(ins.clone());
                 }
