@@ -9,7 +9,7 @@ use cached::Cached;
 use circuit_base::{
     cumulant::{dim_permutation_for_circuits, partitions},
     visit_circuit, Add, CircResult, CircuitNode, CircuitRc, CircuitType, Cumulant, DiscreteVar,
-    Einsum, GeneralFunction, Index, ScalarConstant,
+    Einsum, GeneralFunction, Index, Scalar,
 };
 use circuit_rewrites::circuit_optimizer::{optimize_and_evaluate, OptimizationSettings};
 use itertools::Itertools;
@@ -48,7 +48,7 @@ pub trait SampleSpecTrait {
     fn sample_var(&self, d: &DiscreteVar) -> Result<CircuitRc>;
     fn get_empirical_weights(&self) -> CircuitRc;
     fn get_overall_weight(&self) -> CircuitRc {
-        ScalarConstant::nrc(
+        Scalar::nrc(
             count_elements(&self.get_sample_shape()) as f64,
             Sv::new(),
             None,
@@ -246,7 +246,7 @@ impl SampleSpecTrait for RandomSampleSpec {
         Ok(res)
     }
     fn get_empirical_weights(&self) -> CircuitRc {
-        ScalarConstant::new(
+        Scalar::new(
             1. / count_elements(&self.shape) as f64,
             self.shape.clone(),
             Some("empirical_weights_for_sampled".to_owned()),
@@ -584,7 +584,7 @@ impl SampleTransform {
         matcher: IterativeMatcherRc,
     ) -> Result<CircuitRc> {
         match cum.nodes.len() {
-            0 => Ok(ScalarConstant::new(1., Sv::new(), None).rc()),
+            0 => Ok(Scalar::new(1., Sv::new(), None).rc()),
             1 => {
                 let sampled = self.sample(cum.children().next().unwrap())?;
                 self.get_expectation(sampled)
