@@ -13,7 +13,7 @@ pub fn default_index_traversal() -> IterativeMatcherRc {
             false,
             None,
             None,
-            Matcher::types(vec![CircuitType::Index, CircuitType::ArrayConstant]).rc(),
+            Matcher::types(vec![CircuitType::Index, CircuitType::Array]).rc(),
         )
         .rc()
 }
