@@ -7,8 +7,6 @@ from typing import List, Tuple
 import math
 import itertools
 
-none: None = None
-
 
 def check_type(x, type_v):
     if not isinstance(x, type_v):
@@ -20,14 +18,6 @@ def get_tensor_shape(x):
     return list(x.size())
 
 
-def get_id(x):
-    return id(x)
-
-
-def cast_int(x):
-    return int(x)
-
-
 dtype_from_string = {
     "float32": torch.float32,
     "float64": torch.float64,
@@ -86,6 +76,7 @@ generalfunctions = {
     "q_from_qr": lambda x: torch.linalg.qr(x)[0],
     "min": lambda x: torch.min(x, dim=-1)[0],
     "max": lambda x: torch.max(x, dim=-1)[0],
+    "last_dim_size": lambda x: torch.full(x.shape[:-1], x.shape[-1], dtype=x.dtype, device=x.device),
 }
 
 
@@ -156,8 +147,6 @@ def tensor_from_bytes(device_dtype, shape, bytes, count):
 
 einsum = lambda tensors_and_axes, out_axes: torch.einsum(*itertools.chain(*tensors_and_axes), out_axes)
 
-print = print
-
 
 def random_indices(input: torch.Tensor, shape, replacement: bool) -> torch.Tensor:
     shape = tuple(shape)
