@@ -24,7 +24,6 @@ pub struct PyUtils {
     tensor_scale: PyObject,
     pub generalfunctions: std::collections::HashMap<String, PyObject>,
     pub gen_index_function: PyObject,
-    pub none: PyObject,
     einsum: PyObject,
     make_diagonal: PyObject,
     pub print: PyObject,
@@ -33,6 +32,7 @@ pub struct PyUtils {
     pub tensor_to_bytes: PyObject,
     pub tensor_from_bytes: PyObject,
     pub assert_tensors_close: PyObject,
+    pub builtins: Py<PyModule>,
 }
 
 /// misc python utilities
@@ -49,12 +49,14 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
     .unwrap();
 
     let get = |s: &str| utils.getattr(s).unwrap().into();
+    let builtins = PyModule::import(py, "builtins").unwrap();
+    let get_builtin = |s: &str| builtins.getattr(s).unwrap().into();
 
     PyUtils {
         torch: get("torch"),
         get_tensor_shape: get("get_tensor_shape"),
-        id: get("get_id"),
-        cast_int: get("cast_int"),
+        id: get_builtin("id"),
+        cast_int: get_builtin("int"),
         scalar_to_tensor: get("scalar_to_tensor"),
         cast_tensor: get("cast_tensor"),
         un_flat_concat: get("un_flat_concat"),
@@ -69,15 +71,15 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
             .unwrap()
             .extract()
             .unwrap(),
-        none: get("none"),
         einsum: get("einsum"),
         make_diagonal: get("make_diagonal"),
-        print: get("print"),
+        print: get_builtin("print"),
         random_indices: get("random_indices"),
         conv: get("conv"),
         tensor_to_bytes: get("tensor_to_bytes"),
         tensor_from_bytes: get("tensor_from_bytes"),
         assert_tensors_close: get("assert_tensors_close"),
+        builtins: builtins.into(),
     }
 });
 
