@@ -900,6 +900,27 @@ impl RearrangeSpec {
             .canonicalize(true)
     }
 
+    pub fn forced_input_shape(
+        &self,
+        coerce: bool, /* if true, don't force single int */
+    ) -> OpShape {
+        let mut ret = sv![OpSize::NONE; self.input_ints.len()];
+        for (ints, o) in self.input_ints.iter().zip(&mut ret) {
+            if coerce && ints.len() == 1 {
+                continue;
+            }
+
+            *o = ints
+                .iter()
+                .map(|x| -> Option<usize> { self.int_sizes[*x as usize].into() })
+                .collect::<Option<Vec<_>>>()
+                .map(|x| x.iter().product())
+                .into();
+        }
+
+        ret
+    }
+
     pub fn conform_to_input_shape(
         &self,
         shape: &Shape,
