@@ -2,6 +2,7 @@ use std::{
     cmp::{max, min},
     fmt::{self, Display},
     iter::zip,
+    num::{ParseFloatError, ParseIntError},
     ops::Deref,
 };
 
@@ -690,9 +691,7 @@ impl TensorIndex {
                 .map(str::trim)
                 .filter(|x| !x.is_empty())
                 .map(|x| -> Result<_> {
-                    let num_er = ParseError::NumberFail {
-                        string: x.to_string(),
-                    };
+                    // TODO: fix error handle + trim whitespace
                     if let Ok(int) = x.parse::<i64>() {
                         return Ok(TensorAxisIndex::Single(int));
                     }
@@ -710,18 +709,27 @@ impl TensorIndex {
                                 ))
                             });
                     }
-                    if let [start, end] = x.split(':').collect::<Vec<_>>()[..] {
+
+                    let get_slice_bound = |item: &str, name: &str| {
+                        if item.is_empty() {
+                            Ok(None)
+                        } else {
+                            item.parse::<i64>()
+                                .map_err(ParseError::from)
+                                .with_context(|| {
+                                    format!(
+                                        "failed to parse {} of slice, index string='{}'",
+                                        name, string
+                                    )
+                                })
+                                .map(Some)
+                        }
+                    };
+
+                    if let [start, stop] = x.split(':').map(str::trim).collect::<Vec<_>>()[..] {
                         return Ok(TensorAxisIndex::Slice(Slice {
-                            start: if start.is_empty() {
-                                None
-                            } else {
-                                Some(start.parse::<i64>().map_err(|_e| num_er.clone())?)
-                            },
-                            stop: if end.is_empty() {
-                                None
-                            } else {
-                                Some(end.parse::<i64>().map_err(|_e| num_er.clone())?)
-                            },
+                            start: get_slice_bound(start, "start")?,
+                            stop: get_slice_bound(stop, "stop")?,
                         }));
                     }
                     bail!(er.clone())
@@ -972,32 +980,64 @@ impl Default for TorchDeviceDtype {
     }
 }
 
+const REGEX_101_MSG: &str = "\nTry using regex101.com : )\n(TODO: auto generate regex test page via https://github.com/firasdib/Regex101/wiki/API#python-3)";
+
+// TODO: break up and move part to parsing.rs
 #[apply(python_error_exception)]
 #[base_error_name(Parse)]
 #[base_exception(PyValueError)]
 #[derive(Error, Debug, Clone)]
 pub enum ParseError {
-    #[error("Parsing no regex match '{line}'")]
-    NoRegexMatch { line: String },
+    #[error("regex didn't match '{line}'\nregex='{regex_string}'{}", REGEX_101_MSG)]
+    RegexDidntMatch { line: String, regex_string: String },
 
-    #[error("Parsing number failed {string}")]
-    NumberFail { string: String },
+    #[error(
+        "regex didn't match group={group} for '{line}'\n{}\nregex='{regex_string}'{}",
+        "Note: this error can also be caused by failing to lookup a reference circuit",
+        REGEX_101_MSG
+    )]
+    RegexDidntMatchGroup {
+        line: String,
+        regex_string: String,
+        group: usize,
+    },
+
+    #[error("Parsing number failed err={err_string}")]
+    NumberFail { err_string: String },
+
+    #[error("extra='{extra}'")]
+    ExpectedNoExtraInfo { extra: String },
 
     #[error("Parsing failed {string}")]
     Fail { string: String },
 
-    #[error("Expected UUID string, found  {string}")]
-    InvalidUuid { string: String },
+    #[error("Expected UUID string, found  '{string}'\nerr_msg={err_msg}")]
+    InvalidUuid { string: String, err_msg: String },
 
-    #[error("Parsing wrong number of children, expected {expected} found {found}")]
-    WrongNumberChildren { expected: usize, found: usize },
+    #[error("Parsing wrong number of children, expected {expected} found {found}, ident={ident}")]
+    WrongNumberChildren {
+        expected: usize,
+        found: usize,
+        ident: String,
+    },
 
-    #[error("Parsing invalid indentation, tab width {tab_width} num spaces {spaces} stack indentation {stack_indentation} stack top {stack_top:?}")]
+    #[error("effectively negative indent!\n{}\n{}\n{}",
+        format!("this_num_spaces_base={} < first_num_spaces={}", this_num_spaces_base, first_num_spaces),
+        format!("first_line='{}'", first_line),
+        format!("this_line='{}'", this_line),
+        )]
+    LessIndentationThanFirstItem {
+        this_num_spaces_base: usize,
+        first_num_spaces: usize,
+        first_line: String,
+        this_line: String,
+    },
+
+    #[error("Parsing invalid indentation, tab width {tab_width} num spaces {spaces} stack indentation {stack_indentation}")]
     InvalidIndentation {
         spaces: usize,
         tab_width: usize,
         stack_indentation: usize,
-        stack_top: Option<String>,
     },
 
     #[error("Parsing invalid circuit variant {v}")]
@@ -1012,18 +1052,74 @@ pub enum ParseError {
     #[error("Parsing shape needed but not provided on {variant}")]
     ShapeNeeded { variant: String },
 
-    #[error("Parsing found multiple circuits in sequence, instead of one circuit (aka last node wasn't child of anything)")]
-    GotMultipleCircuits {},
-
     #[error("Circuit by name not found: {name}")]
     CircuitRefNotFound { name: String },
 
-    #[error("Module by name not found: {name}")]
-    ModuleNotFound { name: String },
-
     #[error("Einsum string invalid {string} {substring}")]
     EinsumStringInvalid { string: String, substring: String },
 
     #[error("Einsum string doesn't have arrow {string}")]
     EinsumStringNoArrow { string: String },
+
+    #[error("actual_num_circuits={actual_num_circuits}")]
+    ExpectedOneCircuitGotMultiple { actual_num_circuits: usize },
+
+    #[error("expected 't' or 'f' got='{got}'")]
+    InvalidTerseBool { got: String },
+
+    #[error("reference_name={reference_name}")]
+    ReferenceCircuitHasChildrenInParseStr { reference_name: String },
+
+    #[error("reference_name={reference_name}")]
+    ReferenceCircuitNameFollowedByAdditionalInfo {
+        reference_name: String,
+        line: String,
+    },
+
+    #[error("repeat_ident={repeat_ident} \n{err_strs}line='{line}'")]
+    OnCircuitRepeatInfoIsNotSame {
+        repeat_ident: String,
+        line: String,
+        err_strs: String,
+    },
+
+    #[error("num_chars={num_chars} extra='{extra}' name={name:?}\nmodule extra should look like 'tt,tf,ft' and we check for 2 chars per comma delimited item.")]
+    Expected2CharsPerForModuleExtra {
+        name: Option<String>,
+        num_chars: usize,
+        extra: String,
+    },
+
+    #[error("num_children={num_children} != 2 * num_args={num_args:?} + 1\nname={name:?} extra={extra}\n{}\n{}\n{}",
+    "modules have 1 child for the spec and then a symbol and a input circuit for each arg (2 * num_args + 1)",
+    "this implies the number of children is always odd",
+    "num_args is based on parsing the batchable and expandable info from 'extra' (see above) (in a form like 'tt,tf,ft' (Some(3)) or 'all_t' (None))")]
+    ModuleInconsistentNumChildren {
+        name: Option<String>,
+        num_children: usize,
+        extra: String,
+        num_args: Option<usize>,
+    },
+
+    #[error("factors={factors:?} string='{string}'")]
+    FactorProductTooLarge { factors: Vec<usize>, string: String },
+
+    #[error("i={i} >= bound={bound}")]
+    SymbolicSizeNumberOutOfBounds { i: usize, bound: usize },
+}
+
+impl From<ParseIntError> for ParseError {
+    fn from(value: ParseIntError) -> Self {
+        ParseError::NumberFail {
+            err_string: value.to_string(),
+        }
+    }
+}
+
+impl From<ParseFloatError> for ParseError {
+    fn from(value: ParseFloatError) -> Self {
+        ParseError::NumberFail {
+            err_string: value.to_string(),
+        }
+    }
 }
