@@ -7,6 +7,7 @@ use std::{
 };
 
 use anyhow::{Context, Result};
+use itertools::Itertools;
 use macro_rules_attribute::apply;
 use once_cell::sync::Lazy;
 use pyo3::{types::PyBytes, FromPyObject, IntoPy, PyObject, Python};
@@ -64,6 +65,21 @@ pub fn is_unique<T: Eq + Hash>(col: &[T]) -> bool {
     col.len() == set.len()
 }
 
+pub fn counts<T: Eq + Hash>(x: impl IntoIterator<Item = T>) -> HashMap<T, usize> {
+    let mut map = HashMap::default();
+    for item in x {
+        map.entry(item).and_modify(|x| *x += 1).or_insert(1);
+    }
+    map
+}
+
+pub fn counts_g_1<T: Eq + Hash>(x: impl IntoIterator<Item = T>) -> HashMap<T, usize> {
+    counts(x)
+        .into_iter()
+        .filter_map(|(x, count)| (count > 1).then_some((x, count)))
+        .collect()
+}
+
 pub fn unique_to_appearance<T: Eq + Hash + Clone>(vec: &Vec<T>) -> HashMap<T, Vec<usize>> {
     let mut map: HashMap<T, Vec<usize>> = HashMap::default();
     for (i, x) in vec.iter().enumerate() {
@@ -386,12 +402,8 @@ macro_rules! python_println{
                 s=format!("TRIED TO PRINT TOO LONG {}",&s[..1000]);
             }
             pyo3::Python::with_gil(|py| {
-                py.eval(
-                    &format!("print('''{}''')", s),
-                    None,
-                    None,
-                ).unwrap();
-            })
+                $crate::py_types::PY_UTILS.print.call1(py, (s,)).unwrap()
+            });
         }
     }
 }
@@ -586,3 +598,63 @@ where
         self.0.iter_mut().map(Iterator::next).collect()
     }
 }
+
+#[macro_export]
+macro_rules! fn_struct{
+    ($vis:vis $name:ident : Fn($($iname:ident:$ity:ty),*)->$retty:ty)=>{
+        paste::paste!{
+            #[derive(Clone,Debug)]
+            $vis enum $name{
+                Py(pyo3::PyObject),
+                Dyn([< $name DynStruct >]),
+            }
+            impl $name{
+                fn call(&self,$($iname:$ity),*)->anyhow::Result<$retty>{
+                    match &self{
+                        $name::Py(pyobj)=>rr_util::pycall!(pyobj,($($iname,)*),anyhow),
+                        $name::Dyn(dynf)=>dynf.0($($iname),*),
+                    }
+                }
+                fn new_dyn(f:Box<dyn Fn($($ity),*)->anyhow::Result<$retty> + Send + Sync>)->Self{
+                    Self::Dyn([< $name DynStruct >] (std::sync::Arc::new(f)))
+                }
+            }
+            #[pyclass]
+            #[derive(Clone)]
+            $vis struct [< $name DynStruct >] (
+                std::sync::Arc<dyn Fn($($ity),*)->anyhow::Result<$retty> + Send + Sync>
+            );
+            #[pymethods]
+            impl  [< $name DynStruct >]{
+                fn __call__( &self,
+                    _py: Python<'_>,$($iname:$ity),*)->anyhow::Result<$retty>{
+                        self.0($($iname),*)
+                }
+            }
+            impl<'source> pyo3::FromPyObject<'source> for $name{
+                fn extract(inp: &'source pyo3::PyAny) -> pyo3::PyResult<Self> {
+                    Ok(Self::Py(inp.into()))
+                }
+            }
+            impl IntoPy<PyObject> for $name{
+                fn into_py(self, py: Python<'_>) -> PyObject {
+                    match self{
+                        $name::Py(x)=>x,
+                        $name::Dyn(dynf)=>dynf.into_py(py)
+                    }
+                }
+            }
+            impl Debug for [< $name DynStruct >]{
+                fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+                    write!(f,"dyn_fn_struct")
+                }
+            }
+        }
+    }
+}
+
+pub fn indent(x: String, width: usize) -> String {
+    let tab = " ".repeat(width);
+
+    x.lines().map(|s| tab.clone() + s).join("\n")
+}
