@@ -1,6 +1,7 @@
 use circuit_base::{
     expand_node::ExpandError, generalfunction::GeneralFunctionShapeError,
-    opaque_iterative_matcher::IncorrectOpaqueTypeError, ConstructError, TensorEvalError,
+    opaque_iterative_matcher::IncorrectOpaqueTypeError, parsing::ParseArgError, ConstructError,
+    TensorEvalError,
 };
 use circuit_rewrites::{
     algebraic_rewrite::{DistributeError, PushDownIndexError},
@@ -17,6 +18,7 @@ use pyo3::{
 use rr_util::{
     errors_util::HasPythonException,
     rearrange_spec::{PermError, RearrangeParseError, RearrangeSpecError},
+    symbolic_size::{SymbolicSizeOverflowError, SymbolicSizeSetError},
     tensor_util::{IndexError, MiscInputError, ParseError},
 };
 macro_rules! setup_errors {
@@ -68,5 +70,8 @@ setup_errors!(
     DistributeError,
     SimpConfigError,
     IncorrectOpaqueTypeError,
-    GeneralFunctionShapeError
+    GeneralFunctionShapeError,
+    ParseArgError,
+    SymbolicSizeOverflowError,
+    SymbolicSizeSetError,
 );
