@@ -84,9 +84,9 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     )?;
     m.add_class::<rr_util::opt_einsum::EinsumSpec>()?;
     m.add_class::<circuit_base::Einsum>()?;
-    m.add_class::<circuit_base::ArrayConstant>()?;
+    m.add_class::<circuit_base::Array>()?;
     m.add_class::<circuit_base::Symbol>()?;
-    m.add_class::<circuit_base::ScalarConstant>()?;
+    m.add_class::<circuit_base::Scalar>()?;
     m.add_class::<circuit_base::Add>()?;
     m.add_class::<circuit_base::Rearrange>()?;
     m.add_class::<circuit_base::Index>()?;
@@ -104,10 +104,10 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<circuit_rewrites::scheduled_execution::ScheduleStats>()?;
     m.add_class::<circuit_rewrites::schedule_send::ScheduleToSend>()?;
 
-    m.add_class::<circuit_base::module_nodes::ModuleNode>()?;
-    m.add_class::<circuit_base::module_nodes::ModuleNodeSpec>()?;
-    m.add_class::<circuit_base::module_nodes::ModuleNodeArgSpec>()?;
-    m.add_class::<circuit_base::module_nodes::ModuleNodeArgSpec>()?;
+    m.add_class::<circuit_base::module::Module>()?;
+    m.add_class::<circuit_base::module::ModuleSpec>()?;
+    m.add_class::<circuit_base::module::ModuleArgSpec>()?;
+    m.add_function(wrap_pyfunction!(circuit_base::module::all_free, m)?)?;
     m.add_class::<circuit_base::Tag>()?;
     m.add_class::<circuit_base::DiscreteVar>()?;
     m.add_class::<circuit_base::StoredCumulantVar>()?;
@@ -130,6 +130,8 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
 
     m.add_class::<rr_util::lru_cache::TensorCacheRrfs>()?;
 
+    m.add_function(wrap_pyfunction!(rr_util::symbolic_size::symbolic_sizes, m)?)?;
+
     m.add_function(wrap_pyfunction!(
         circuit_rewrites::module_rewrite::fuse_concat_modules,
         m
@@ -145,6 +147,9 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     )?)?;
     m.add_function(wrap_pyfunction!(circuit_base::circuit_is_var_py, m)?)?;
 
+    m.add_function(wrap_pyfunction!(circuit_base::check_evaluable, m)?)?;
+    m.add_function(wrap_pyfunction!(circuit_base::evaluate_device_dtype, m)?)?;
+
     m.add_function(wrap_pyfunction!(add_collapse_scalar_inputs, m)?)?;
     m.add_function(wrap_pyfunction!(add_deduplicate, m)?)?;
     m.add_function(wrap_pyfunction!(remove_add_few_input, m)?)?;
@@ -218,6 +223,9 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         m
     )?)?;
 
+    m.add_function(wrap_pyfunction!(nb_operations::diff::diff_circuits, m)?)?;
+    m.add_function(wrap_pyfunction!(nb_operations::diff::compute_self_hash, m)?)?;
+
     m.add_function(wrap_pyfunction!(index_einsum_to_scatter, m)?)?;
     m.add_function(wrap_pyfunction!(scatter_elim_identity, m)?)?;
     m.add_function(wrap_pyfunction!(einsum_pull_scatter, m)?)?;
@@ -325,7 +333,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     )?)?;
     m.add_function(wrap_pyfunction!(rr_util::rrfs::tensor_from_hash, m)?)?;
     m.add_function(wrap_pyfunction!(
-        circuit_base::module_nodes::inline_all_modules,
+        circuit_base::module::inline_all_modules,
         m
     )?)?;
     m.add_function(wrap_pyfunction!(
@@ -366,10 +374,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         circuit_rewrites::server::circuit_server_serve,
         m
     )?)?;
-    m.add_function(wrap_pyfunction!(
-        circuit_base::parsing::parse_circuit_py,
-        m
-    )?)?;
+    m.add_class::<circuit_base::parsing::Parser>()?;
 
     m.add_function(wrap_pyfunction!(circuit_base::print_circuit_type_check, m)?)?;
     m.add_function(wrap_pyfunction!(
