@@ -1,6 +1,6 @@
 use std::fs;
 
-use circuit_base::parsing::parse_circuit;
+use circuit_base::parsing::Parser;
 use circuit_rewrites::circuit_optimizer::{
     optimize_circuit, OptimizationContext, OptimizationSettings,
 };
@@ -16,14 +16,15 @@ fn benchmarking_circs() {
         let circuits: Vec<_> = fs::read_dir(format!("{}/ryan/compiler_benches/", get_rrfs_dir()))
             .unwrap()
             .filter_map(|d| {
-                parse_circuit(
-                    &fs::read_to_string(d.unwrap().path()).unwrap(),
+                Parser::new(
                     Default::default(),
                     Default::default(),
                     true,
                     Default::default(),
-                    &mut None,
+                    true,
                 )
+                .unwrap()
+                .parse_circuit(&fs::read_to_string(d.unwrap().path()).unwrap(), &mut None)
                 .ok()
             })
             .collect();
